import pytest
import numpy as np
import qp
from tests.helpers.test_data_helper import NPDF


def test_create(hist_test_data, norm_test_data):
    """Make sure that qp create works when the actual class is passed."""

    test_data = hist_test_data["hist"]["ctor_data"]
    ens_h = qp.create(qp.hist, test_data)

    assert ens_h.npdf == NPDF
    assert ens_h.metadata["pdf_name"][0].decode() == "hist"

    test_data_n = norm_test_data["norm"]["ctor_data"]
    ens_n = qp.create(qp.stats.norm, test_data_n)
    assert ens_n.metadata["pdf_name"][0].decode() == "norm"


@pytest.mark.parametrize(
    "filename, expected",
    [("test-quant.h5", True), ("test-quant-nocheckinput.h5", False)],
)
def test_qp_read_files_with_check_input(filename, expected, test_data_dir):
    """Make sure that qp read works with Ensembles with the check_input parameter."""

    filepath = test_data_dir / filename
    ens_q = qp.read(filepath)

    assert ens_q.metadata["ensure_extent"] == expected


def test_qp_read_with_fmt(test_data_dir):
    """Make sure that qp read works when given the fmt argument"""

    filepath = test_data_dir / "test.hdf5"
    ens = qp.read(filepath, fmt="hdf5")

    assert ens.metadata["pdf_name"][0].decode() == "mixmod"


@pytest.mark.parametrize(
    "ancil1,ancil2,ancil3",
    [
        ({"ids": np.arange(0, 1)}, {"ids": np.arange(1, 5)}, {"ids": np.arange(5, 13)}),
        (
            {"ids": np.array([[0]])},
            {"ids": np.array([[1], [2], [3], [4]])},
            {"ids": np.array([[5], [6], [7], [8], [9], [10], [11], [12]])},
        ),
    ],
)
def test_ensemble_concatenation(ancil1, ancil2, ancil3, hist_ensemble):
    """Tests that ensemble concatenation works with 1D and 2D ancillary data, and
    ensembles with 1 or more distributions."""

    # create 3 ensembles with ancillary data

    ens_1 = hist_ensemble[0:1]
    ens_2 = hist_ensemble[1:5]
    ens_3 = hist_ensemble[1:9]

    ens_1.set_ancil(ancil1)
    ens_2.set_ancil(ancil2)
    ens_3.set_ancil(ancil3)
    new_ens = qp.concatenate([ens_1, ens_2, ens_3])

    assert new_ens.npdf == ens_1.npdf + ens_2.npdf + ens_3.npdf
    assert len(new_ens.ancil["ids"]) == new_ens.npdf



import pytest
import warnings
import numpy as np
import qp

from tests.helpers.test_data_helper import NPDF
from tests.helpers.test_funcs import assert_all_close


def test_norm(hist_test_data):
    """Test that the histogram norm method works as expected."""

    norm = False
    ancil = {"ancil": np.linspace(0, 11, NPDF)}
    ens_h = qp.hist.create_ensemble(
        **hist_test_data["hist"]["ctor_data"], norm=norm, ancil=ancil
    )

    assert ens_h.npdf == NPDF

    # make sure that values were not normalized
    xmin = ens_h.frozen.dist._xmin
    xmax = ens_h.frozen.dist._xmax
    integral = ens_h.integrate([xmin, xmax])
    assert not np.allclose(integral, np.ones(NPDF))

    # make sure that the normalization has worked as expected
    ens_h.norm()
    integral_norm = ens_h.integrate([xmin, xmax])
    assert_all_close(integral_norm, np.ones(NPDF))

    # make sure that the ancil table has not changed
    assert ens_h.ancil == ancil


def test_xsamples(hist_ensemble):
    """Make sure that x_samples generates values as expected."""

    x_samps = hist_ensemble.x_samples()
    assert len(hist_ensemble.metadata["bins"]) - 1 == len(x_samps)

    bin_centers = (
        hist_ensemble.metadata["bins"][:-1] + hist_ensemble.metadata["bins"][1:]
    ) / 2

    assert_all_close(x_samps, bin_centers)


@pytest.mark.parametrize(
    "bins, pdfs",
    [
        (np.array([0, 0.5, 1, np.inf]), np.array([1.0, 2.0, 3.0])),
        (np.array([0, 0.5, 1, 1.5]), np.array([1, np.nan, 1])),
        (np.array([0, 0.5, 1, 1.5]), np.array([-0.5, 1, 1.5])),
    ],
)
def test_warnings(bins, pdfs):
    """Test that warnings when building a histogram Ensemble work."""

    with pytest.warns(RuntimeWarning):
        ens_h = qp.hist.create_ensemble(bins=bins, pdfs=pdfs)


@pytest.mark.parametrize(
    "bins,pdfs, match_string",
    [
        (
            np.array([1.0, 0.5, 1.5, 2]),
            np.array([1, 2, 3]),
            "Invalid bins: The given bins are not sorted:",
        ),
        (np.array([0, 0.5, 1, 1.5]), np.array([1, 2, 3, 4]), "Number of bins"),
        (
            np.array([0, 0.5, 1, 1.5]),
            np.array([-0.5, -1, -0.5]),
            "cannot be properly normalized",
        ),
    ],
)
def test_invalid_input(bins, pdfs, match_string):
    """Tests that the appropriate errors are raised when invalid input is given"""

    with pytest.raises(ValueError, match=match_string) as err:
        ens_h = qp.hist.create_ensemble(bins=bins, pdfs=pdfs, warn=False)


def test_hist_bins2d():
    """Test that passing 2D bin array works"""

    bins = np.array([[0, 0.5, 1, 1.5]])
    pdfs = np.array([1, 2, 3])

    ens = qp.hist.create_ensemble(bins=bins, pdfs=pdfs)



"""This module implements a distribution parameterization sub-class using interpolated grids"""

from __future__ import annotations
import numpy as np
from scipy.stats import rv_continuous
from typing import Mapping, Optional
from numpy.typing import ArrayLike
import warnings

from .interp_utils import (
    irreg_interp_extract_xy_vals,
    extract_vals_at_x,
    extract_xy_sparse,
)
from ...core.factory import add_class
from ...core.ensemble import Ensemble
from ..base import Pdf_rows_gen
from ...plotting import get_axes_and_xlims, plot_pdf_on_axes
from ...utils.array import reshape_to_pdf_size, reduce_dimensions
from ...utils.interpolation import (
    interpolate_multi_x_multi_y,
    interpolate_multi_x_y,
    interpolate_x_multi_y,
)


class interp_gen(Pdf_rows_gen):
    """Implements distributions parameterized as interpolated sets of values.

    All distributions share the same x values. Interpolation is performed using
    `scipy.interpolate.interp1d`, with the default interpolation method (linear).


    Parameters
    ----------
    xvals : ArrayLike
        The n x-values that are used by all the distributions
    yvals : ArrayLike
        The y-values that represent each distribution, with shape (npdf,n)
    norm : bool, optional
        If `True`, normalizes the input distribution. If `False`, assumes the
        given distribution is already normalized. By default `True`.
    warn : bool, optional
        If `True`, raises warnings if input is not valid PDF data (i.e. if
        data is negative). If `False`, no warnings are raised. By default `True`.


    Notes
    -----

    Converting to this parameterization:

    This table contains the available methods to convert to this parameterization,
    their required arguments, and their method keys. If the key is `None`, this is
    the default conversion method.

    +---------------------+-----------+------------+
    | Function            | Arguments | Method key |
    +---------------------+-----------+------------+
    | `.extract_vals_at_x`| xvals     | None       |
    +---------------------+-----------+------------+

    Implementation notes:

    This uses the same xvals for all the the PDFs, unlike `interp_irregular_gen` which
    has a different set of xvals for each distribution.
    `interp_gen` therefore allows for much faster evaluation than `interp_irregular_gen`,
    and reduces the memory usage by a factor of 2.

    Inside the range of given xvals it takes a set of x and y values
    and uses `scipy.interpolate.interp1d` to build the PDF.
    Outside the range of given xvals the `pdf()` will return 0.

    The `cdf()` is constructed by integrating analytically -- computing the cumulative
    sum at the given xvals and interpolating between them.
    This will give a slight discrepancy with the true integral of the `pdf()`,
    but is much, much faster to evaluate.
    Outside the range of given xvals the `cdf()` will return 0 or 1, respectively

    The `ppf()` is computed by inverting the `cdf()`. `ppf(0)` will return negative infinity
    and `ppf(1)` will return positive infinity.
    """

    # pylint: disable=protected-access

    name = "interp"
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(
        self,
        xvals: ArrayLike,
        yvals: ArrayLike,
        norm: bool = True,
        warn: bool = True,
        *args,
        **kwargs,
    ):
        """
        Create a new distribution by interpolating the given values

        Parameters
        ----------
        xvals : ArrayLike
            The n x-values that are used by all the distributions
        yvals : ArrayLike
            The y-values that represent each distribution, with shape (npdf,n)
        norm : bool, optional
            If `True`, normalizes the input distribution. If `False`, assumes the
            given distribution is already normalized. By default `True`.
        warn : bool, optional
            If True, raises warnings if input is not valid PDF data (i.e. if
            data is negative). If `False`, no warnings are raised. By default `True`.
        """
        if np.size(xvals) != np.shape(yvals)[-1]:  # pragma: no cover
            raise ValueError(
                "Shape of xbins in xvals (%s) != shape of xbins in yvals (%s)"
                % (np.size(xvals), np.shape(yvals)[-1])
            )
        self._xvals = np.asarray(xvals)
        self._yvals = reshape_to_pdf_size(np.asarray(yvals), -1)
        kwargs["shape"] = np.shape(self._yvals)

        # make sure that the xvals are sorted
        if not np.all(np.diff(self._xvals) >= 0):
            raise ValueError(
                f"Invalid xvals: The given xvals are not sorted: {self._xvals}"
            )

        # raise warnings if input data is not finite or pdfs are not positive
        self._warn = warn
        if self._warn:
            if not np.all(np.isfinite(self._xvals)):
                warnings.warn(
                    "The given xvals contain non-finite values", RuntimeWarning
                )
            if not np.all(np.isfinite(self._yvals)):
                indices = np.where(np.isfinite(self._yvals) != True)
                warnings.warn(
                    f"There are non-finite values in the yvals for the following distributions: {indices}",
                    RuntimeWarning,
                )
            if np.any(self._yvals < 0):
                indices = np.where(self._yvals < 0)
                warnings.warn(
                    f"There are negative values in the yvals for the following distributions: {indices}",
                    RuntimeWarning,
                )

        # Set support
        self._xmin = self._xvals[0]
        self._xmax = self._xvals[-1]
        # kwargs["shape"] = np.shape(yvals)

        # normalize the distribution if norm is True
        self._norm = norm
        if self._norm:
            self._yvals = self.normalize()["yvals"]
        else:  # pragma: no cover
            self._ycumul = None

        super().__init__(*args, **kwargs)
        self._addmetadata("xvals", self._xvals)
        self._addobjdata("yvals", self._yvals)

    def _compute_ycumul(self) -> None:
        """Compute the integral under the distribution"""
        copy_shape = np.array(self._yvals.shape)
        self._ycumul = np.ndarray(copy_shape)
        self._ycumul[:, 0] = 0.5 * self._yvals[:, 0] * (self._xvals[1] - self._xvals[0])
        self._ycumul[:, 1:] = np.cumsum(
            (self._xvals[1:] - self._xvals[:-1])
            * 0.5
            * np.add(self._yvals[:, 1:], self._yvals[:, :-1]),
            axis=1,
        )

        # raise an error if the sum is 0 or negative
        if np.any(self._ycumul[:, -1] < 0):
            indices = np.where(self._ycumul[:, -1] < 0)
            raise ValueError(
                f"The distribution(s) cannot be properly normalized, the integral is < 0 for distributions at indices = {indices[0]} \n with yvals {self._yvals[indices[0]]}"
            )
        elif np.any(self._ycumul[:, -1] == 0):
            indices = np.where(self._ycumul[:, -1] == 0)
            warnings.warn(
                f"The distributions at indices = {indices[0]} have an integral of 0."
            )

    def normalize(self) -> Mapping[str, np.ndarray[float]]:
        """Normalizes the input distribution values.

        Returns
        -------
        Mapping [str, np.ndarray[float]]
            An (npdf, n) array of y values for the npdf distributions

        Raises
        ------
        ValueError
            Raised if the sum under the distribution <= 0.
        """

        self._compute_ycumul()

        new_yvals = (self._yvals.T / self._ycumul[:, -1]).T
        self._ycumul = (self._ycumul.T / self._ycumul[:, -1]).T
        return {"yvals": new_yvals}

    @property
    def xvals(self) -> np.ndarray[float]:
        """Return the x-values used to do the interpolation"""
        return self._xvals

    @property
    def yvals(self) -> np.ndarray[float]:
        """Return the y-values used to do the interpolation"""
        return self._yvals

    def x_samples(self) -> np.ndarray[float]:
        """Return a set of x values that can be used to plot all the PDFs."""
        return self._xvals

    def _pdf(self, x, row):
        # pylint: disable=arguments-differ
        pdf = interpolate_x_multi_y(
            x, row, self._xvals, self._yvals, bounds_error=False, fill_value=0.0
        ).ravel()

        # reduce dimension to 0 if there's only one value
        if np.shape(pdf) == (1,) and len(pdf) == 1:
            return pdf[0]
        else:
            return pdf

    def _cdf(self, x, row):
        # pylint: disable=arguments-differ
        if self._ycumul is None:  # pragma: no cover
            self._compute_ycumul()
        return interpolate_x_multi_y(
            x, row, self._xvals, self._ycumul, bounds_error=False, fill_value=(0.0, 1.0)
        ).ravel()

    def _ppf(self, x, row):
        # pylint: disable=arguments-differ
        if self._ycumul is None:  # pragma: no cover
            self._compute_ycumul()

        return interpolate_multi_x_y(
            x,
            row,
            self._ycumul,
            self._xvals,
            bounds_error=False,
            fill_value=(self._xmin, self._xmax),
        ).ravel()

    def _munp(self, m, *args):
        """compute moments"""
        # pylint: disable=arguments-differ
        # Silence floating point warnings from integration.
        with np.errstate(all="ignore"):
            vals = self.custom_generic_moment(m)
        return vals

    def custom_generic_moment(self, m):
        """Compute the mth moment"""
        m = np.asarray(m)
        dx = self._xvals[1] - self._xvals[0]
        return np.sum(self._xvals**m * self._yvals, axis=1) * dx

    def _updated_ctor_param(self):
        """
        Sets the arguments as additional constructor arguments. This function is needed
        by scipy in order to copy distributions, and makes a dictionary of all parameters
        necessary to construct the distribution.
        """
        dct = super()._updated_ctor_param()
        dct["xvals"] = self._xvals
        dct["yvals"] = self._yvals
        dct["norm"] = self._norm
        dct["warn"] = self._warn
        return dct

    @classmethod
    def get_allocation_kwds(
        cls, npdf: int, **kwargs
    ) -> dict[str, tuple[tuple[int, int], str]]:
        """Return the kwds necessary to create an `empty` HDF5 file with ``npdf`` entries
        for iterative write. We only need to allocate the data columns, as
        the metadata will be written when we finalize the file.

        The number of data columns is calculated based on the length or shape of the
        metadata, ``n``. For example, the number of columns is ``nbins-1``
        for a histogram.

        Parameters
        ----------
        npdf : int
            Total number of distributions that will be written out
        kwargs :
            The keys needed to construct the shape of the data to be written.

        Returns
        -------
        dict[str, tuple[tuple[int, int], str]]
            A dictionary with a key for the objdata, a tuple with the shape of that data,
            and the data type of the data as a string.
            i.e. ``{objdata_key = ( (npdf, n), "f4" )}``

        Raises
        ------
        ValueError
            Raises an error if xvals is not provided.
        """
        if "xvals" not in kwargs:  # pragma: no cover
            raise ValueError("required argument xvals not included in kwargs")
        ngrid = np.shape(kwargs["xvals"])[-1]
        return dict(yvals=((npdf, ngrid), "f4"))

    @classmethod
    def plot_native(cls, pdf, **kwargs):
        """Plot the PDF in a way that is particular to this type of distribution

        For a interpolated PDF this uses the interpolation points.

        Parameters
        ----------
        axes : Axes
            The axes to plot on. Either this or xlim must be provided.
        xlim : tuple[float, float]
            The x-axis limits. Either this or axes must be provided.
        kwargs :
            Any keyword arguments to pass to `matplotlib.axes.Axes.hist`.

        Returns
        -------
        axes : Axes
            The plot axes.
        """
        axes, _, kw = get_axes_and_xlims(**kwargs)
        return plot_pdf_on_axes(axes, pdf, pdf.dist.xvals, **kw)

    @classmethod
    def add_mappings(cls) -> None:
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)
        cls._add_extraction_method(extract_vals_at_x, None)

    @classmethod
    def create_ensemble(
        self,
        xvals: ArrayLike,
        yvals: ArrayLike,
        norm: bool = True,
        warn: bool = True,
        ancil: Optional[Mapping] = None,
    ) -> Ensemble:
        """Creates an Ensemble of distributions parameterized as interpolations.


        Parameters
        ----------
        xvals : ArrayLike
          The x-values used to do the interpolation, shape is n
        yvals : ArrayLike
          The y-values used to do the interpolation, shape is (npdfs, n), where
          npdfs is the number of distributions
        norm : bool, optional
            If True, normalizes the input distribution. If False, assumes the
            given distribution is already normalized. By default True.
        warn : bool, optional
            If True, raises warnings if input is not valid PDF data (i.e. if
            data is negative). If False, no warnings are raised. By default True.
        ancil : Optional[Mapping]
            A dictionary of metadata for the distributions, where any arrays
            have the same length as the number of distributions

        Returns
        -------
        Ensemble
            An Ensemble object containing all of the given distributions.

        Examples
        --------

        To create an ensemble with two distributions and their associated ids:

        >>> import qp
        >>> import numpy as np
        >>> xvals= np.array([0,0.5,1,1.5,2]),
        >>> yvals = np.array([[0.01, 0.2,0.3,0.2,0.01],[0.09,0.25,0.2,0.1,0.01]])
        >>> ancil = {'ids':[5,8]}
        >>> ens = qp.interp.create_ensemble(xvals, yvals,ancil=ancil)
        >>> ens.metadata
        {'pdf_name': array([b'interp'], dtype='|S6'),
        'pdf_version': array([0]),
        'xvals': array([[0. , 0.5, 1. , 1.5, 2. ]])}

        """
        data = {"xvals": xvals, "yvals": yvals, "norm": norm, "warn": warn}
        return Ensemble(self, data, ancil)


interp = interp_gen


class interp_irregular_gen(Pdf_rows_gen):
    """Implements distributions parameterized as interpolated sets of values.

    Each distribution has its own set of x values. Interpolation is performed using
    `scipy.interpolate.interp1d`, with the default interpolation method (linear).

    Parameters
    ----------
    xvals : ArrayLike
        The x-values that are used by each distribution, with shape (npdf,n)
    yvals : ArrayLike
        The y-values that represent each distribution, with shape (npdf,n)
    norm : bool, optional
        If True, normalizes the input distribution. If False, assumes the
        given distribution is already normalized. By default True.
    warn : bool, optional
        If True, raises warnings if input is not valid PDF data (i.e. if
        data is negative). If False, no warnings are raised. By default True.


    Notes
    -----

    Converting to this parameterization:

    This table contains the available methods to convert to this parameterization,
    their required arguments, and their method keys. If the key is `None`, this is
    the default conversion method.

    +-------------------------------------+-----------+------------+
    | Function                            | Arguments | Method key |
    +-------------------------------------+-----------+------------+
    | `.irreg_interp_extract_xy_vals`     | xvals     | None       |
    +-------------------------------------+-----------+------------+

    Implementation notes:

    Inside the range xvals[:,0], xvals[:,-1] it simply takes a set of x and y values
    and uses `scipy.interpolate.interp1d` to linearly interpolate the PDF.
    Outside the range xvals[:,0], xvals[:,-1] the `pdf()` will return 0.

    The cdf() is constructed by analytically computing the cumulative
    sum at the xvals grid points and linearly interpolating between them.
    This will give a slight discrepancy with the true integral of the `pdf()`,
    but is much, much faster to evaluate.
    Outside the range xvals[:,0], xvals[:,-1] the `cdf()` will return 0 or 1, respectively

    The `ppf()` is computed by inverting the `cdf()`. `ppf(0)` gives negative infinity, and
    `ppf(1)` gives positive infinity.

    """

    # pylint: disable=protected-access

    name = "interp_irregular"
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(
        self,
        xvals: ArrayLike,
        yvals: ArrayLike,
        norm: bool = True,
        warn: bool = True,
        *args,
        **kwargs,
    ):
        """
        Create a new distribution by interpolating the given values

        Parameters
        ----------
        xvals : ArrayLike
          The x-values for each distribution, with shape (npdf, n), where n is
          the number of x-values
        yvals : ArrayLike
          The y-values that represent each distribution, with shape (npdf,n)
        norm : bool, optional
            If True, normalizes the input distribution. If False, assumes the
            given distribution is already normalized. By default True.
        warn : bool, optional
            If True, raises warnings if input is not valid PDF data (i.e. if
            data is negative). If False, no warnings are raised. By default True.
        """
        if np.shape(xvals) != np.shape(yvals):  # pragma: no cover
            raise ValueError(
                "Shape of xvals (%s) != shape of yvals (%s)"
                % (np.shape(xvals), np.shape(yvals))
            )
        self._xvals = reshape_to_pdf_size(np.asarray(xvals), -1)
        self._yvals = reshape_to_pdf_size(np.asarray(yvals), -1)

        # raise warnings if input data is not finite or pdfs are not positive
        self._warn = warn
        if self._warn:
            if not np.all(np.isfinite(self._xvals)):
                indices = np.where(np.isfinite(xvals) != True)
                warnings.warn(
                    f"The given xvals contain non-finite values for the following distributions: {indices}",
                    RuntimeWarning,
                )
            if not np.all(np.isfinite(self._yvals)):
                indices = np.where(np.isfinite(self._yvals) != True)
                warnings.warn(
                    f"There are non-finite values in the yvals for the following distributions: {indices}",
                    RuntimeWarning,
                )
            if np.any(self._yvals < 0):
                indices = np.where(self._yvals < 0)
                warnings.warn(
                    f"There are negative values in the yvals for the following distributions: {indices}",
                    RuntimeWarning,
                )

        # make sure that the xvals are sorted
        if not np.all(np.diff(self._xvals) >= 0):
            raise ValueError(
                f"Invalid xvals: The given xvals are not sorted: {self._xvals}"
            )

        self._xmin = np.min(self._xvals)
        self._xmax = np.max(self._xvals)
        # kwargs["shape"] = np.shape(xvals)[:-1]
        kwargs["shape"] = np.shape(self._xvals)

        self._norm = norm

        if self._norm:
            self._yvals = self.normalize()["yvals"]
        self._ycumul = None
        super().__init__(*args, **kwargs)
        self._addobjdata("xvals", self._xvals)
        self._addobjdata("yvals", self._yvals)

    def _compute_ycumul(self):
        copy_shape = np.array(self._yvals.shape)
        self._ycumul = np.ndarray(copy_shape)
        self._ycumul[:, 0] = 0.0
        self._ycumul[:, 1:] = np.cumsum(
            self._xvals[:, 1:] * self._yvals[:, 1:]
            - self._xvals[:, :-1] * self._yvals[:, 1:],
            axis=1,
        )

        # make sure that integrals are > 0
        if np.any(self._ycumul[:, -1] < 0):
            indices = np.where(self._ycumul[:, -1] < 0)
            raise ValueError(
                f"The integral is < 0 for distributions at indices = {indices[0]}, so the distribution(s) cannot be properly normalized."
            )
        elif np.any(self._ycumul[:, -1] == 0):
            indices = np.where(self._ycumul[:, -1] == 0)
            warnings.warn(
                f"The distributions at indices = {indices[0]} have an integral of 0."
            )

    def normalize(self) -> Mapping[str, np.ndarray[float]]:
        """
        Normalize a set of 1D interpolators

        Returns
        -------
        ynorm : Mapping[str, np.ndarray[float]]
            Normalized y-vals
        """
        # def row_integral(irow):
        #    return quad(interp1d(xvals[irow], yvals[irow], **kwargs), limits[0], limits[1])[0]

        # vv = np.vectorize(row_integral)
        # integrals = vv(np.arange(xvals.shape[0]))
        integrals = np.sum(
            self._xvals[:, 1:] * self._yvals[:, 1:]
            - self._xvals[:, :-1] * self._yvals[:, 1:],
            axis=1,
        )

        # make sure that integrals are >= 0
        if np.any(integrals < 0):
            indices = np.where(integrals < 0)
            raise ValueError(
                f"The integral is < 0 for distributions at indices = {indices[0]}, so the distribution(s) cannot be properly normalized."
            )
        elif np.any(integrals == 0):
            indices = np.where(integrals == 0)
            warnings.warn(
                f"The distributions at indices = {indices[0]} have an integral of 0."
            )

        return {"yvals": (self._yvals.T / integrals).T}

    @property
    def xvals(self) -> np.ndarray[float]:
        """Return the x-values used to do the interpolation"""
        return self._xvals

    @property
    def yvals(self) -> np.ndarray[float]:
        """Return the y-valus used to do the interpolation"""
        return self._yvals

    def x_samples(self) -> np.ndarray[float]:
        """Return a set of x values that can be used to plot all the PDFs."""
        dx = np.min(np.diff(self._yvals))
        return np.arange(np.min(self._xvals), np.max(self._yvals), dx)

    def _pdf(self, x, row):
        # pylint: disable=arguments-differ
        pdf = interpolate_multi_x_multi_y(
            x, row, self._xvals, self._yvals, bounds_error=False, fill_value=0.0
        ).ravel()

        # reduce dimension to 0 if there's only one value
        if np.shape(pdf) == (1,) and len(pdf) == 1:
            return pdf[0]
        else:
            return pdf

    def _cdf(self, x, row):
        # pylint: disable=arguments-differ
        if self._ycumul is None:  # pragma: no cover
            self._compute_ycumul()
        return interpolate_multi_x_multi_y(
            x, row, self._xvals, self._ycumul, bounds_error=False, fill_value=(0.0, 1.0)
        ).ravel()

    def _ppf(self, x, row):
        # pylint: disable=arguments-differ
        if self._ycumul is None:  # pragma: no cover
            self._compute_ycumul()
        return interpolate_multi_x_multi_y(
            x,
            row,
            self._ycumul,
            self._xvals,
            bounds_error=False,
            fill_value=(self._xmin, self._xmax),
        ).ravel()

    def _updated_ctor_param(self):
        """
        Set the bins as additional constructor argument
        """
        dct = super()._updated_ctor_param()
        dct["xvals"] = self._xvals
        dct["yvals"] = self._yvals
        dct["norm"] = self._norm
        dct["warn"] = self._warn
        return dct

    @classmethod
    def get_allocation_kwds(
        cls, npdf: int, **kwargs
    ) -> dict[str, tuple[tuple[int, int], str]]:
        """Return the kwds necessary to create an `empty` HDF5 file with ``npdf`` entries
        for iterative write. We only need to allocate the data columns, as
        the metadata will be written when we finalize the file.

        The number of data columns is calculated based on the length or shape of the
        metadata, ``n``. For example, the number of columns is ``nbins-1``
        for a histogram.

        Parameters
        ----------
        npdf : int
            Total number of distributions that will be written out
        kwargs :
            The keys needed to construct the shape of the data to be written.

        Returns
        -------
        dict[str, tuple[tuple[int, int], str]]
            A dictionary with a key for the objdata, a tuple with the shape of that data,
            and the data type of the data as a string.
            i.e. ``{objdata_key = ((npdf, n), "f4")}``

        Raises
        ------
        ValueError
            Raises an error if xvals is not provided.
        """
        if "xvals" not in kwargs:  # pragma: no cover
            raise ValueError("required argument xvals not included in kwargs")
        ngrid = np.shape(kwargs["xvals"])[-1]
        return dict(xvals=((npdf, ngrid), "f4"), yvals=((npdf, ngrid), "f4"))

    @classmethod
    def plot_native(cls, pdf, **kwargs):
        """Plot the PDF in a way that is particular to this type of distribution

        For a interpolated PDF this uses the interpolation points.

        Parameters
        ----------
        axes : Axes
            The axes to plot on. Either this or xlim must be provided.
        xlim : tuple[float, float]
            The x-axis limits. Either this or axes must be provided.
        kwargs :
            Any keyword arguments to pass to matplotlib's axes.hist() method.

        Returns
        -------
        axes : Axes
            The plot axes.
        """
        axes, _, kw = get_axes_and_xlims(**kwargs)
        xvals_row = pdf.dist.xvals
        return plot_pdf_on_axes(axes, pdf, xvals_row, **kw)

    @classmethod
    def add_mappings(cls) -> None:
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)
        cls._add_extraction_method(irreg_interp_extract_xy_vals, None)
        cls._add_extraction_method(extract_xy_sparse, "sparse")

    @classmethod
    def create_ensemble(
        self,
        xvals: ArrayLike,
        yvals: ArrayLike,
        norm: bool = True,
        warn: bool = True,
        ancil: Optional[Mapping] = None,
    ) -> Ensemble:
        """Creates an Ensemble of distributions parameterized as interpolations.

        Parameters
        ----------
        xvals : ArrayLike
          The x-values for each distribution, with shape (npdf, n), where n is
          the number of x-values
        yvals : ArrayLike
          The y-values that represent each distribution, with shape (npdf,n)
        norm : bool, optional
            If True, normalizes the input distribution. If False, assumes the
            given distribution is already normalized. By default True.
        warn : bool, optional
            If True, raises warnings if input is not valid PDF data (i.e. if
            data is negative). If False, no warnings are raised. By default True.
        ancil : Optional[Mapping]
            A dictionary of metadata for the distributions, where any arrays have the same length as the number of distributions.

        Returns
        -------
        Ensemble
            An Ensemble object containing all of the given distributions.

        Examples
        --------

        To create an Ensemble with two distributions and their associated ids:

        >>> import qp
        >>> import numpy as np
        >>> xvals = np.array([[0,0.5,1,1.5,2],[0.5,0.75,1,1.25,1.5]]),
        >>> yvals = np.array([[0.01, 0.2,0.3,0.2,0.01],[0.09,0.25,0.2,0.1,0.01]])}
        >>> ancil = {'ids':[5,8]}
        >>> ens = qp.interp_irregular.create_ensemble(xvals, yvals,ancil)
        >>> ens.metadata
        {'pdf_name': array([b'interp_irregular'], dtype='|S16'),
        'pdf_version': array([0])}

        """
        data = {"xvals": xvals, "yvals": yvals, "norm": norm, "warn": warn}
        return Ensemble(self, data, ancil)


interp_irregular = interp_irregular_gen
add_class(interp_gen)
add_class(interp_irregular_gen)



from __future__ import annotations

import sys
import numpy as np
from numpy.typing import ArrayLike

from ...utils.array import (
    get_eval_case,
    CASE_2D,
    CASE_FLAT,
    CASE_FACTOR,
    CASE_PRODUCT,
    get_bin_indices,
)

# Conversion functions


def extract_quantiles(in_dist: "Ensemble", **kwargs) -> dict[str, np.ndarray[float]]:
    """Convert using a set of quantiles and the locations at which they are reached

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions

    Other Parameters
    ----------------
    quants : np.ndarray
        Quantile values to use

    Returns
    -------
    data : dict[str, Any]
        The extracted data
    """
    quants = kwargs.pop("quants", None)
    if quants is None:  # pragma: no cover
        raise ValueError("To convert using extract_quantiles you must specify quants")
    locs = in_dist.ppf(quants)
    return dict(quants=quants, locs=locs)


# Creation functions


def pad_quantiles(
    quants: ArrayLike, locs: ArrayLike
) -> tuple[np.ndarray[float], np.ndarray[float]]:
    """Pad the quantiles and locations used to build a quantile representation.
    Ensuring 0 and 1 are part of quantiles.
    Extrapolates loc at 0 by taking a linear extrapolation from the first two points
    and following to where it intersects 0
    Extrapolates loc at 1 by taking a linear extrapolation from the last two points
    and following to where it intersects 1

    This will add additional data points to the quants and locs


    Parameters
    ----------
    quants : ArrayLike
        The quantiles used to build the CDF
    locs : ArrayLike
        The locations at which those quantiles are reached

    Returns
    -------
    quants : np.ndarray[float]
        The quantiles used to build the CDF
    locs : np.ndarray[float]
        The locations at which those quantiles are reached
    """
    n_out = n_vals = quants.size
    if quants[0] > sys.float_info.epsilon:
        offset_lo = 1
        pad_lo = True
        n_out += 1
    else:
        offset_lo = 0
        pad_lo = False
    if quants[-1] < 1.0:
        pad_hi = True
        n_out += 1
    else:
        pad_hi = False
    if n_out == n_vals:
        return quants, locs
    quants_out = np.zeros((n_out), quants.dtype)
    locs_out = np.zeros((locs.shape[0], n_out), quants.dtype)
    quants_out[offset_lo : n_vals + offset_lo] = quants
    locs_out[:, offset_lo : n_vals + offset_lo] = locs
    if pad_lo:
        locs_out[:, 0] = locs[:, 0] - quants[0] * (locs[:, 1] - locs[:, 0]) / (
            quants[1] - quants[0]
        )

    if pad_hi:
        quants_out[-1] = 1.0
        locs_out[:, -1] = locs[:, -1] - (1.0 - quants[-1]) * (
            locs[:, -2] - locs[:, -1]
        ) / (quants[-1] - quants[-2])

    return quants_out, locs_out


def evaluate_hist_multi_x_multi_y(
    x: ArrayLike, row: ArrayLike, bins: ArrayLike, vals: ArrayLike, derivs=None
) -> np.ndarray[float]:
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : ArrayLike
        X values to interpolate at
    row : ArrayLike
        Which rows to interpolate at
    bins : ArrayLike, shape (npdf, N+1)
        'x' bin edges
    vals : ArrayLike, shape (npdf, N)
        'y' bin contents

    Returns
    -------
    out : np.ndarray[float]
        The histogram values
    """
    case_idx, xx, rr = get_eval_case(x, row)
    if case_idx in [CASE_PRODUCT, CASE_FACTOR]:
        return evaluate_hist_multi_x_multi_y_product(xx, rr, bins, vals, derivs)
    if case_idx == CASE_2D:
        return evaluate_hist_multi_x_multi_y_2d(xx, rr, bins, vals, derivs)
    return evaluate_hist_multi_x_multi_y_flat(xx, rr, bins, vals, derivs)


def evaluate_hist_multi_x_multi_y_flat(
    x: ArrayLike, row: ArrayLike, bins: ArrayLike, vals: ArrayLike, derivs=None
) -> np.ndarray[float]:  # pragma: no cover
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : ArrayLike, length n
        X values to interpolate at
    row : ArrayLike, length n
        Which rows to interpolate at
    bins : ArrayLike, shape (npdf, N+1)
        'x' bin edges
    vals : ArrayLike, shape (npdf, N)
        'y' bin contents

    Returns
    -------
    out : np.ndarray[float], length n
        The histogram values
    """

    def evaluate_row(xv, rv):
        bins_row = bins[rv]
        idx, mask = get_bin_indices(bins_row, xv)
        delta = xv - bins_row[idx]
        if derivs is None:
            return np.where(mask, vals[rv, idx], 0)
        return np.where(mask, vals[rv, idx] + delta * derivs[rv, idx], 0)

    vv = np.vectorize(evaluate_row)
    return vv(x, row)


def evaluate_hist_multi_x_multi_y_product(
    x: ArrayLike, row: ArrayLike, bins: ArrayLike, vals: ArrayLike, derivs=None
) -> np.ndarray[float]:  # pragma: no cover
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : ArrayLike, length npts
        X values to interpolate at
    row : ArrayLike, shape (npdf, 1)
        Which rows to interpolate at
    bins : ArrayLike, shape (npdf, N+1)
        'x' bin edges
    vals : ArrayLike, shape (npdf, N)
        'y' bin contents

    Returns
    -------
    out : np.ndarray[float], shape (npdf, npts)
        The histogram values
    """

    def evaluate_row(rv):
        bins_flat = bins[rv].flatten()
        idx, mask = get_bin_indices(bins_flat, x)
        delta = x - bins_flat[idx]
        if derivs is None:
            return np.where(mask, np.squeeze(vals[rv])[idx], 0).flatten()
        return np.where(
            mask, np.squeeze(vals[rv])[idx] + delta * np.squeeze(derivs[rv])[idx], 0
        )

    vv = np.vectorize(evaluate_row, signature="(1)->(%i)" % (x.size))
    return vv(row)


def evaluate_hist_multi_x_multi_y_2d(
    x: ArrayLike, row: ArrayLike, bins: ArrayLike, vals: ArrayLike, derivs=None
) -> np.ndarray[float]:  # pragma: no cover
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : ArrayLike, shape (npdf, npts)
        X values to interpolate at
    row : ArrayLike, shape (npdf, 1)
        Which rows to interpolate at
    bins : ArrayLike, shape (npdf, N+1)
        'x' bin edges
    vals : ArrayLike, shape (npdf, N)
        'y' bin contents

    Returns
    -------
    out : np.ndarray[float], shape (npdf, npts)
        The histogram values
    """
    nx = np.shape(x)[-1]

    def evaluate_row(rv, xv):
        flat_bins = bins[rv].flatten()
        idx, mask = get_bin_indices(flat_bins, xv)
        delta = xv - flat_bins[idx]
        if derivs is None:
            return np.where(mask, np.squeeze(vals[rv])[idx], 0).flatten()
        return np.where(
            mask, np.squeeze(vals[rv])[idx] + delta * np.squeeze(derivs[rv])[idx], 0
        ).flatten()

    vv = np.vectorize(evaluate_row, signature="(1),(%i)->(%i)" % (nx, nx))
    return vv(row, x)



from __future__ import annotations  # for autodoc type annotations
import numpy as np

from ...utils.array import (
    get_bin_indices,
    get_eval_case,
    CASE_PRODUCT,
    CASE_FACTOR,
    CASE_2D,
)
from typing import Any

from numpy.typing import ArrayLike


#
# PDF evaluation functions
#


def evaluate_hist_x_multi_y(
    x: ArrayLike, row: ArrayLike, bins: ArrayLike, vals: ArrayLike, derivs=None
) -> np.ndarray[float]:
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : ArrayLike
        X values to interpolate at
    row : ArrayLike
        Which rows to interpolate at
    bins : ArrayLike, length N+1
        'x' bin edges
    vals : ArrayLike, shape (npdf,N)
        'y' bin contents

    Returns
    -------
    out : np.ndarray[float]
        The histogram values

    Notes
    -----
    Depending on the shape of 'x' and 'row' this will
    use one of the three specific implementations.
    """
    case_idx, xx, rr = get_eval_case(x, row)
    if case_idx in [CASE_PRODUCT, CASE_FACTOR]:
        return evaluate_hist_x_multi_y_product(xx, rr, bins, vals, derivs)
    if case_idx == CASE_2D:
        return evaluate_hist_x_multi_y_2d(xx, rr, bins, vals, derivs)
    return evaluate_hist_x_multi_y_flat(xx, rr, bins, vals, derivs)


def evaluate_hist_x_multi_y_product(
    x: ArrayLike, row: ArrayLike, bins: ArrayLike, vals: ArrayLike, derivs=None
) -> np.ndarray[float]:  # pragma: no cover
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : ArrayLike, length npts
        X values to interpolate at
    row : ArrayLike, shape (npdf, 1)
        Which rows to interpolate at
    bins : ArrayLike, length (N+1)
        'x' bin edges
    vals : ArrayLike, shape (npdf, N)
        'y' bin contents

    Returns
    -------
    out : np.ndarray[float], shape (npdf, npts)
        The histogram values
    """
    # assert np.ndim(x) < 2 and np.ndim(row) == 2
    idx, mask0 = get_bin_indices(bins, x)
    mask = np.ones(row.shape) * mask0
    if derivs is None:
        return np.where(mask, vals[:, idx][np.squeeze(row)], 0)
    deltas = x - bins[idx]
    return np.where(
        mask,
        vals[:, idx][np.squeeze(row)] + deltas * derivs[:, idx][np.squeeze(row)],
        0,
    )


def evaluate_hist_x_multi_y_2d(
    x: ArrayLike, row: ArrayLike, bins: ArrayLike, vals: ArrayLike, derivs=None
) -> np.ndarray[float]:  # pragma: no cover
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : ArrayLike, shape (npdf, npts)
        X values to interpolate at
    row : ArrayLike, shape (npdf, 1)
        Which rows to interpolate at
    bins : ArrayLike, length (N+1)
        'x' bin edges
    vals : ArrayLike, shape (npdf, N)
        'y' bin contents

    Returns
    -------
    out : np.ndarray[float], shape (npdf, npts)
        The histogram values
    """
    assert np.ndim(x) >= 2 and np.ndim(row) >= 2
    idx, mask = get_bin_indices(bins, x)
    if derivs is None:
        deltas = np.zeros(idx.shape)
    else:
        deltas = x - bins[idx]

    def evaluate_row(idxv, maskv, rv, delta):
        if derivs is None:
            return np.where(maskv, vals[rv, idxv], 0)
        return np.where(maskv, vals[rv, idxv] + delta * derivs[rv, idxv], 0)

    vv = np.vectorize(evaluate_row)
    return vv(idx, mask, row, deltas)


def evaluate_hist_x_multi_y_flat(
    x: ArrayLike, row: ArrayLike, bins: ArrayLike, vals: ArrayLike, derivs=None
) -> np.ndarray[float]:  # pragma: no cover
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : ArrayLike, length n
        X values to interpolate at
    row : ArrayLike, length n
        Which rows to interpolate at
    bins : ArrayLike, length N+1
        'x' bin edges
    vals : ArrayLike, shape (npdf, N)
        'y' bin contents

    Returns
    -------
    out : np.ndarray[float], length n
        The histogram values
    """
    assert np.ndim(x) < 2 and np.ndim(row) < 2
    idx, mask = get_bin_indices(bins, x)
    if derivs is None:
        deltas = np.zeros(idx.shape)
    else:
        deltas = x - bins[idx]

    def evaluate_row(idxv, maskv, rv, delta):
        if derivs is None:
            return np.where(maskv, vals[rv, idxv], 0)
        return np.where(maskv, vals[rv, idxv] + delta * derivs[rv, idxv], 0)

    vv = np.vectorize(evaluate_row)
    return vv(idx, mask, row, deltas)


#
# Conversion functions
#


def extract_hist_values(
    in_dist: "Ensemble", **kwargs
) -> dict[str, Any]:  # -> dict[str, Any]:
    """Convert to a histogram by using the CDF values at the given bin edges
    to calculate the value within each bin.

    Parameters
    ----------
    in_dist : Ensemble
        Input Ensemble of distributions.

    Other Parameters
    ----------------
    bins : np.ndarray[float]
        The bin edges for the new histogram.

    Returns
    -------
    data : dict[str, Any]
        The extracted data
    """
    bins = kwargs.pop("bins", None)
    if bins is None:  # pragma: no cover
        raise ValueError("To convert using extract_hist_samples you must specify bins")
    bins, pdfs = in_dist.histogramize(bins)
    return dict(bins=bins, pdfs=pdfs)


def extract_hist_samples(in_dist: "Ensemble", **kwargs) -> dict[str, Any]:
    """Convert to a histogram by sampling the input distributions, then turning
    those samples into a histogram with given bin edges.

    Parameters
    ----------
    in_dist : Ensemble
        Input Ensemble of distributions

    Other Parameters
    ----------------
    bins : np.ndarray[float]
        The bin edges for the new histogram
    size : int
        Number of samples to generate, by default 1000

    Returns
    -------
    data : dict[str, Any]
        The extracted data
    """
    bins = kwargs.pop("bins", None)
    size = kwargs.pop("size", 1000)
    if bins is None:  # pragma: no cover
        raise ValueError("To convert using extract_hist_samples you must specify bins")
    samples = in_dist.rvs(size=size)

    def hist_helper(sample):
        return np.histogram(sample, bins=bins)[0]

    vv = np.vectorize(
        hist_helper, signature="(%i)->(%i)" % (samples.shape[0], bins.size - 1)
    )
    pdfs = vv(samples)
    return dict(bins=bins, pdfs=pdfs)



import sys
import os

# Provide path to the python modules we want to run autodoc on
sys.path.insert(0, os.path.abspath("../qp"))

import qp

# set up readthedocs theme
import sphinx_rtd_theme

html_theme = "sphinx_rtd_theme"


# Avoid imports that may be unsatisfied when running sphinx, see:
# http://stackoverflow.com/questions/15889621/sphinx-how-to-exclude-imports-in-automodule#15912502
autodoc_mock_imports = ["scipy", "scipy.interpolate", "sklearn"]

# set up extensions

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.doctest",
    "sphinx_design",
    "myst_nb",
    "sphinx_copybutton",
    "sphinx.ext.intersphinx",
    "numpydoc",
]

myst_enable_extensions = [
    "colon_fence",
    "dollarmath",
    "attrs_inline",
    "tasklist",
    "deflist",
]
myst_heading_anchors = 5
nb_execution_mode = "auto"
nb_execution_allow_errors = True
exclude_patterns = [
    "_build",
    "_build/jupyter_execute",
    "_build/html/_downloads",
]  # where not to render notebook files to markdown
copybutton_exclude = ".linenos, .gp"  # what to exclude from copied code blocks

# set up autodocs
master_doc = "index"
autosummary_generate = True
autoclass_content = "class"
autodoc_default_flags = ["members", "no-special-members"]
autodoc_member_order = "bysource"
autodoc_type_aliases = {
    # requires `from __future__ import annotations` in all modules
    "ArrayLike": "~numpy.typing.ArrayLike",  # required to stop ArrayLike from expanding in the function signature, can't seem to make it link unfortunately
}

numpydoc_show_class_members = False
numpydoc_validation_checks = {"SS01"}
numpydoc_xref_param_type = True  # save from having to backtick everything
numpydoc_xref_aliases = {
    "Optional": "typing.Optional",  # required for links in docstrings
    "Mapping": "typing.Mapping",  # required for links in docstrings
    "Union": "typing.Union",  # required for links in docstrings
    "Any": "typing.Any",  # required for links in docstrings
    "List": "typing.List",  # required for links in docstrings
    # external
    "ArrayLike": "numpy.typing.ArrayLike",  # required for links in docstrings
    "Axes": "matplotlib.axes.Axes",
    "Figure": "matplotlib.figure.Figure",
    "sps.gaussian_kde": "scipy.stats.gaussian_kde ",
    "np.uint8": "numpy.uint8",
    # qp classes: lets you write just "ClassName" in files other than where ClassName is defined
    # otherwise you'd need to write what's defined below
    "Pdf_gen": "`~qp.parameterizations.base.Pdf_gen`",
    "Ensemble": "`~qp.Ensemble`",
    "AbstractQuantilePdfConstructor": "`~qp.parameterizations.quant.abstract_pdf_constructor.AbstractQuantilePdfConstructor`",
}
numpydoc_xref_ignore = {
    "optional",
    "default",
    "subclass",
    "of",
    "or",
    "length",
    "size",
    "shape",
    "containing",
    "distributions",
}


# intersphinx configuration
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "h5py": ("https://docs.h5py.org/en/stable", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
}
default_role = "py:obj"  # interpret `function` as crossref to the py object 'function'


# html_sidebars = {
#     "**": ["globaltoc.html", "relations.html", "sourcelink.html", "searchbox.html"],
# }
html_static_path = ["_static"]
html_css_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.1.1/css/all.min.css",
    "custom.css",
]
html_theme_options = {"style_external_links": True}

project = "qp"
author = "DESC RAIL Team"
copyright = "2025, " + author
version = qp.__version__
release = qp.__version__



from mpi4py import MPI
import qp

# get the rank and size from MPI
rank = MPI.COMM_WORLD.rank
size = MPI.COMM_WORLD.size

# choose the chunk size for each iteration
chunk_size = 5

# the path to the file to iterate through
file_path = "./test.hdf5"

# iterate through the file
for start, end, ens_chunk in qp.iterator(
    file_path, chunk_size=chunk_size, rank=rank, parallel_size=size
):
    print(f"For rank: {rank}")
    print(f"Indices are: ({start}, {end})")
    print(ens_chunk)



import pytest
import qp
import numpy as np


@pytest.mark.parametrize(
    "data_dir",
    [
        ({"test": np.array(["one", "two"]), "test2": np.array([1, 2.5])}),
        ({"test": ["gal1", "gal2"]}),
        ({"test": np.array(["galaxy1"]), "test2": np.array([0])}),
    ],
)
def test_decode_strings(data_dir):
    """Test that decode strings works with multiple input types."""

    # encode strings
    data_dir["test"] = np.strings.encode(data_dir["test"], "utf-8")
    decoded = qp.array.decode_strings(data_dir)

    assert decoded["test"].dtype.kind == "U"
    if "test2" in data_dir.keys():
        assert np.allclose(decoded["test2"], data_dir["test2"])


def test_decode_strings_not_numpy():
    """Make sure that decode_strings works on something that's not a np.ndarray."""

    uncoded_strs = ["gal1", "gal2"]
    test_strs = {"test": ["gal1".encode(), "gal2".encode()]}
    decoded = qp.array.decode_strings(test_strs)

    for i in range(0, len(uncoded_strs)):
        assert isinstance(decoded["test"][i], str)
        assert uncoded_strs[i] == decoded["test"][i]



from __future__ import annotations

import numpy as np

from ...core.lazy_modules import mixture


def extract_mixmod_fit_samples(
    in_dist: "Ensemble", **kwargs
) -> dict[str, np.ndarray[float]]:
    """Convert to a mixture model using a set of values sampled from the pdf

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions

    Other Parameters
    ----------------
    ncomps : int
        Number of components in mixture model to use
    nsamples : int
        Number of samples to generate
    random_state : int
        Used to reproducibly generate random variate from in_dist

    Returns
    -------
    data : dict[str, np.ndarray[float]]
        The extracted data
    """
    n_comps = kwargs.pop("ncomps", 3)
    n_sample = kwargs.pop("nsamples", 1000)
    random_state = kwargs.pop("random_state", None)
    samples = in_dist.rvs(size=n_sample, random_state=random_state)

    def mixmod_helper(samps):
        estimator = mixture.GaussianMixture(n_components=n_comps)
        estimator.fit(samps.reshape(-1, 1))
        weights = estimator.weights_
        means = estimator.means_[:, 0]
        stdevs = np.sqrt(estimator.covariances_[:, 0, 0])
        ov = np.vstack([weights, means, stdevs])
        return ov

    vv = np.vectorize(mixmod_helper, signature="(%i)->(3,%i)" % (n_sample, n_comps))
    fit_vals = vv(samples)
    return dict(
        weights=fit_vals[:, 0, :], means=fit_vals[:, 1, :], stds=fit_vals[:, 2, :]
    )



"""This module implements a distribution parameterization sub-class using interpolated quantiles"""

from __future__ import annotations
import logging
import sys

import numpy as np
from scipy.stats import rv_continuous
from typing import Mapping, Optional
from numpy.typing import ArrayLike
import warnings

from .quant_utils import extract_quantiles, pad_quantiles
from ...core.factory import add_class
from ...core.ensemble import Ensemble
from ..base import Pdf_rows_gen
from ...plotting import get_axes_and_xlims, plot_pdf_quantiles_on_axes
from . import (
    AbstractQuantilePdfConstructor,
    CdfSplineDerivative,
    DualSplineAverage,
    PiecewiseConstant,
    PiecewiseLinear,
)
from ...utils.array import reshape_to_pdf_size
from ...utils.interpolation import interpolate_multi_x_y, interpolate_x_multi_y

epsilon = sys.float_info.epsilon


DEFAULT_PDF_CONSTRUCTOR = "piecewise_linear"
PDF_CONSTRUCTORS = {
    "cdf_spline_derivative": CdfSplineDerivative,
    "dual_spline_average": DualSplineAverage,
    "piecewise_linear": PiecewiseLinear,
    "piecewise_constant": PiecewiseConstant,
}


class quant_gen(Pdf_rows_gen):  # pylint: disable=too-many-instance-attributes
    """Quantile based distribution, where the PDF is defined from the quantiles.


    Parameters
    ----------
    quants : ArrayLike
        The quantiles of the CDF, of shape n
    locs : ArrayLike
        The locations at which those quantiles are reached, of shape (npdf, n)
    pdf_constructor_name : str, optional
        The constructor or interpolator to use to create the PDF, by default "piecewise_linear".
    ensure_extent : bool, optional
        If True, will ensure that the quants start at 0 and end at 1 by adding
        data points at both ends until this is true. locs are extrapolated linearly
        from input data. By default True.
    warn : bool, optional
        If True, raises warnings if input is not valid data (i.e. if
        data is not finite). If False, no warnings are raised. By default True.


    Notes
    -----

    Converting to this parameterization:

    This table contains the available methods to convert to this parameterization,
    their required arguments, and their method keys. If the key is `None`, this is
    the default conversion method.

    +---------------------+-----------+------------+
    | Function            | Arguments | Method key |
    +---------------------+-----------+------------+
    |`.extract_quantiles` | quants    | None       |
    +---------------------+-----------+------------+

    Implementation notes:

    This implements a CDF by interpolating a set of quantile values

    It takes a set of quants and locs values and uses `scipy.interpolate.interp1d`
    with a spline interpolation method of order 2 (kind=`quadratic`) to build the CDF.

    It has multiple PDF constructors to get the PDF from the quantiles. The default
    is the `piecewise_linear` method, which takes the numerical derivative of the
    CDF and interpolates between those points.

    `ppf(0)` returns negative infinity and `ppf(1)` returns positive infinity.

    """

    # pylint: disable=protected-access

    name = "quant"
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(
        self,
        quants: ArrayLike,
        locs: ArrayLike,
        pdf_constructor_name: str = DEFAULT_PDF_CONSTRUCTOR,
        ensure_extent: bool = True,
        warn: bool = True,
        *args,
        **kwargs,
    ):
        """
        Create a new distribution using the given values

        Parameters
        ----------
        quants : ArrayLike
           The quantiles of the CDF, of shape n
        locs : ArrayLike
           The locations at which those quantiles are reached, of shape (npdf, n)
        pdf_constructor_name : str, optional
            The constructor to use to create the PDF, by default "piecewise_linear".
        ensure_extent : bool, optional
            If True, will ensure that the quants start at 0 and end at 1 by adding
            data points at both ends until this is true. locs are extrapolated linearly
            from input data. By default True.
        warn : bool, optional
            If True, raises warnings if input is not valid data (i.e. if
            data is not finite). If False, no warnings are raised. By default True.
        """

        self._xmin = np.min(locs)
        self._xmax = np.max(locs)

        locs_2d = reshape_to_pdf_size(np.asarray(locs), -1)

        # make sure input makes sense for a CDF
        self._validate_input(np.asarray(quants), locs_2d)

        # check locs are finite
        self._warn = warn
        if self._warn:
            if not np.all(np.isfinite(locs_2d)):
                indices = np.where(np.isfinite(locs_2d) != True)
                warnings.warn(
                    f"There are non-finite values in the locs for the distributions: {indices[0]}",
                    RuntimeWarning,
                )

        self._ensure_extent = ensure_extent
        if self._ensure_extent:
            quants, locs_2d = pad_quantiles(quants, locs_2d)

        self._quants = np.asarray(quants)
        self._nquants = self._quants.size
        if locs_2d.shape[-1] != self._nquants:  # pragma: no cover
            raise ValueError(
                "Number of locations (%i) != number of quantile values (%i)"
                % (self._nquants, locs_2d.shape[-1])
            )
        self._locs = locs_2d

        # set up PDF constructor
        if not isinstance(pdf_constructor_name, str):
            try:
                pdf_constructor_name = str(np.strings.decode(pdf_constructor_name))
            except AttributeError as a_err:
                pdf_constructor_name = str(pdf_constructor_name)

        if pdf_constructor_name not in PDF_CONSTRUCTORS:
            raise ValueError(
                f"Unknown interpolator provided: '{pdf_constructor_name}'. Allowed interpolators are {list(PDF_CONSTRUCTORS.keys())}"  # pylint: disable=line-too-long
            )
        self._pdf_constructor_name = pdf_constructor_name
        self._pdf_constructor = None
        self._instantiate_pdf_constructor()

        kwargs["shape"] = self._locs.shape  # locs.shape
        super().__init__(*args, **kwargs)

        self._addmetadata("quants", self._quants)
        self._addmetadata("pdf_constructor_name", self._pdf_constructor_name.encode())
        self._addmetadata("ensure_extent", self._ensure_extent)
        self._addobjdata("locs", self._locs)

    def _validate_input(self, quants, locs):
        """Ensures that given input matches criteria for a valid CDF."""

        if np.any(quants < 0) or np.any(quants > 1):
            raise ValueError(
                f"Invalid quants: One or more of the given quants is outside the allowed range (0,1): {quants}"
            )
        if not np.all(np.diff(quants) >= 0):
            raise ValueError(
                f"Invalid quants: \n There are decreasing values, quants must be given in order from 0 to 1: {quants}"
            )
        if not np.all(np.diff(locs) >= 0):
            indices = np.where(np.diff(locs) < 0)
            raise ValueError(
                f"Invalid locs: \n The given data does not produce a one-to-one CDF for the distributions at the following indices: {indices}"
            )

    @property
    def quants(self) -> np.ndarray[float]:
        """Return quantiles used to build the CDF"""
        return self._quants

    @property
    def locs(self) -> np.ndarray[float]:
        """Return the locations at which those quantiles are reached"""
        return self._locs

    @property
    def pdf_constructor_name(self) -> str:
        """Returns the name of the current pdf constructor. Matches a key in
        the `PDF_CONSTRUCTORS` dictionary."""
        return self._pdf_constructor_name

    @pdf_constructor_name.setter
    def pdf_constructor_name(self, value: str) -> None:
        """Allows users to specify a different interpolator without having to recreate
        the ensemble.

        Parameters
        ----------
        value : str
            One of the supported interpolators. See `PDF_CONSTRUCTORS`
            dictionary for supported interpolators.

        Raises
        ------
        ValueError
            If the value provided isn't a key in `PDF_CONSTRUCTORS`, raise
            a value error.
        """
        if value not in PDF_CONSTRUCTORS:
            raise ValueError(
                f"Unknown interpolator provided: '{value}'. Allowed interpolators are {list(PDF_CONSTRUCTORS.keys())}"  # pylint: disable=line-too-long
            )

        if value is self._pdf_constructor_name:
            logging.warning("Already using interpolator: '%s'.", value)
            return

        self._pdf_constructor_name = value
        self._instantiate_pdf_constructor()
        self._addmetadata("pdf_constructor_name", self._pdf_constructor_name)

    @property
    def pdf_constructor(self) -> AbstractQuantilePdfConstructor:
        """Returns the current PDF constructor, and allows the user to interact
        with its methods.

        Returns
        -------
        AbstractQuantilePdfConstructor
            Abstract base class of the active concrete PDF constructor.
        """
        return self._pdf_constructor

    def _instantiate_pdf_constructor(self):
        self._pdf_constructor = PDF_CONSTRUCTORS[self._pdf_constructor_name](
            self._quants, self._locs
        )

    def x_samples(self) -> np.ndarray[float]:
        """Return a set of x values that can be used to plot all the CDFs."""

        # get the range and median distance between points
        min_dx = np.median(np.diff(self._locs))
        min_val = np.min(self._locs)
        max_val = np.max(self._locs)

        # get the number of points (make sure it's less than some huge number)
        npts = (max_val - min_val) // min_dx
        npts = np.min([int(npts), 10000])
        return np.linspace(min_val, max_val, npts)

    def _pdf(self, x, *args):
        # We're not requiring that the output be normalized!
        # `util.normalize_interp1d` addresses _one_ of the ways that a reconstruction
        # can be bad, but not all. It should be replaced with a more comprehensive
        # normalization function.
        # See qp issue #147
        row = args[0]
        return self._pdf_constructor.construct_pdf(x, row)

    def _cdf(self, x, row):
        # pylint: disable=arguments-differ
        return interpolate_multi_x_y(
            x,
            row,
            self._locs,
            self._quants,
            bounds_error=False,
            fill_value=(0.0, 1),
            kind="quadratic",
        ).ravel()

    def _ppf(self, x, row):
        # pylint: disable=arguments-differ
        return interpolate_x_multi_y(
            x,
            row,
            self._quants,
            self._locs,
            bounds_error=False,
            fill_value=(self._xmin, self._xmax),
            kind="quadratic",
        ).ravel()

    def _updated_ctor_param(self):
        """
        Set the quants and locs as additional constructor arguments
        """
        dct = super()._updated_ctor_param()
        dct["quants"] = self._quants
        dct["locs"] = self._locs
        dct["pdf_constructor_name"] = self._pdf_constructor_name
        dct["ensure_extent"] = self._ensure_extent
        dct["warn"] = self._warn
        return dct

    @classmethod
    def get_allocation_kwds(
        cls, npdf, **kwargs
    ) -> dict[str, tuple[tuple[int, int], str]]:
        """Return the kwds necessary to create an `empty` HDF5 file with ``npdf`` entries
        for iterative write. We only need to allocate the data columns, as
        the metadata will be written when we finalize the file.

        The number of data columns is calculated based on the length or shape of the
        metadata, ``n``. For example, the number of columns is ``nbins-1``
        for a histogram.

        Parameters
        ----------
        npdf : int
            Total number of distributions that will be written out
        kwargs :
            The keys needed to construct the shape of the data to be written.

        Returns
        -------
        dict[str, tuple[tuple[int, int], str]]
            A dictionary with a key for the objdata, a tuple with the shape of that data,
            and the data type of the data as a string.
            i.e. ``{objdata_key = ((npdf, n), "f4")}``

        Raises
        ------
        ValueError
            Raises an error if the required kwarg quants is not provided.
        """
        try:
            quants = kwargs["quants"]
        except ValueError:  # pragma: no cover
            print("required argument 'quants' not included in kwargs")
        nquants = np.shape(quants)[-1]
        return dict(locs=((npdf, nquants), "f4"))

    @classmethod
    def plot_native(cls, pdf, **kwargs):
        """Plot the PDF in a way that is particular to this type of distribution

        For a quantile this shows the quantiles points.

        Parameters
        ----------
        axes : Axes
            The axes to plot on. Either this or xlim must be provided.
        xlim : tuple[float, float]
            The x-axis limits. Either this or axes must be provided.

        Other Parameters
        ----------------
        npts : int, optional
            The number of x values to create within the limits, by default 101
        kwargs :
            Any keyword arguments to pass to matplotlib's axes.hist() method.

        Returns
        -------
        axes : Axes
            The plot axes.
        """
        axes, xlim, kw = get_axes_and_xlims(**kwargs)
        xvals = np.linspace(xlim[0], xlim[1], kw.pop("npts", 101))
        locs = np.squeeze(pdf.dist.locs[pdf.kwds["row"]])
        quants = np.squeeze(pdf.dist.quants)
        yvals = np.squeeze(pdf.pdf(xvals))
        return plot_pdf_quantiles_on_axes(
            axes, xvals, yvals, quantiles=(quants, locs), **kw
        )

    @classmethod
    def add_mappings(cls) -> None:
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)
        cls._add_extraction_method(extract_quantiles, None)

    @classmethod
    def create_ensemble(
        self,
        quants: ArrayLike,
        locs: ArrayLike,
        pdf_constructor_name: str = DEFAULT_PDF_CONSTRUCTOR,
        ensure_extent: bool = True,
        warn: bool = True,
        ancil: Optional[Mapping] = None,
    ) -> Ensemble:
        """Creates an Ensemble of distributions parameterized as quantiles.


        The options for pdf_constructor_name are: `piecewise_linear`, `piecewise_constant`,
        `dual_spline_average` and 'cdf_spline_derivative`.


        Parameters
        ----------
        quants : ArrayLike
           The quantiles used to build the CDF, shape n
        locs : ArrayLike
           The locations at which those quantiles are reached, shape (npdfs, n),
           where npdfs is the number of distributions.
        pdf_constructor_name : str, optional
            The constructor to use to create the PDF, by default "piecewise_linear".
        ensure_extent : bool, optional
            If True, will ensure that the quants start at 0 and end at 1 by adding
            data points at both ends until this is true. locs are extrapolated linearly
            from input data. By default True.
        warn : bool, optional
            If True, raises warnings if input is not valid (i.e. if
            locs are not finite values). If False, no warnings are raised.
            By default True.
        ancil : Optional[Mapping], optional
            A dictionary of metadata for the distributions, where any arrays have
            the same length as the number of distributions, by default None

        Returns
        -------
        Ensemble
            An Ensemble object containing all of the given distributions.

        Examples
        --------

        To create an Ensemble with two distributions and associated ids, using the
        `dual_spline_average` constructor:

        >>> import qp
        >>> import numpy as np
        >>> quants = np.array([0.0001,0.25,0.5,0.75,0.9999])
        >>> locs = np.array([[0.0001,0.1,0.3,0.5,0.75],[0.01,0.05,0.15,0.3,0.5]])
        >>> pdf_constructor_name = 'dual_spline_average'
        >>> ancil = {'ids':[11,18]}
        >>> ens = qp.quant.create_ensemble(quants,locs,pdf_constructor_name,ancil=ancil)
        >>> ens.metadata
        {'pdf_name': array([b'quant'], dtype='|S5'),
        'pdf_version': array([0]),
        'quants': array([[0.000e+00, 1.000e-04, 2.500e-01, 5.000e-01, 7.500e-01, 9.999e-01,
                1.000e+00]]),
        'pdf_constructor_name': array(['dual_spline_average'], dtype='|S19'),
        'check_input': array([ True])}
        """
        data = {
            "quants": quants,
            "locs": locs,
            "pdf_constructor_name": pdf_constructor_name,
            "ensure_extent": ensure_extent,
            "warn": warn,
        }
        return Ensemble(self, data, ancil)


quant = quant_gen


add_class(quant_gen)



"""This module implements functions to convert distributions between various representations
These functions should then be registered with the `qp.ConversionDict` using `qp_add_mapping`.
That will allow the automated conversion mechanisms to work.
"""

from __future__ import annotations

import numpy as np
from scipy import integrate as sciint
from scipy import interpolate as sciinterp

from ..parameterizations.sparse_interp.sparse_rep import (
    build_sparse_representation,
    indices2shapes,
)


def extract_xy_vals(in_dist: "Ensemble", xvals: np.ndarray) -> dict[str, np.ndarray]:
    """Convert using a set of x and y values.

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions
    xvals : np.ndarray
        Locations at which the pdf is evaluated

    Returns
    -------
    data : dict[str, np.ndarray]
        The extracted data
    """

    yvals = in_dist.pdf(xvals)
    expand_x = np.ones(yvals.shape) * np.squeeze(xvals)
    return dict(xvals=expand_x, yvals=yvals)


#
# Unused conversion functions -- not tested, use at own risk
#


def extract_fit(in_dist, **kwargs):  # pragma: no cover
    """Convert to a functional distribution by fitting it to a set of x and y values

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions

    Other Parameters
    ----------------
    xvals : `np.array`
        Locations at which the pdf is evaluated

    Returns
    -------
    data : dict
        The extracted data
    """
    raise NotImplementedError("extract_fit")
    # xvals = kwargs.pop('xvals', None)
    # if xvals is None:
    #   raise ValueError("To convert using extract_fit you must specify xvals")
    ##vals = in_dist.pdf(xvals)


def extract_voigt_mixmod(in_dist, **kwargs):  # pragma: no cover
    """Convert to a voigt mixture model starting with a gaussian mixture model,
    trivially by setting gammas to 0

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions

    Returns
    -------
    data : dict
        The extracted data
    """
    objdata = in_dist.objdata
    means = objdata["means"]
    stds = objdata["stds"]
    weights = objdata["weights"]
    gammas = np.zeros_like(means)
    return dict(means=means, stds=stds, weights=weights, gammas=gammas, **kwargs)


def extract_voigt_xy(in_dist, **kwargs):  # pragma: no cover
    """Build a voigt function basis and run a match-pursuit algorithm to fit gridded data

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions

    Returns
    -------
    data : dict
        The extracted data as sparse indices, basis, and metadata to rebuild the basis
    """

    sparse_results = extract_voigt_xy_sparse(in_dist, **kwargs)
    indices = sparse_results["indices"]
    meta = sparse_results["metadata"]

    w, m, s, g = indices2shapes(indices, meta)
    return dict(means=m, stds=s, weights=w, gammas=g)


def extract_voigt_xy_sparse(in_dist, **kwargs):  # pragma: no cover
    """Build a voigt function basis and run a match-pursuit algorithm to fit gridded data

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions

    Returns
    -------
    data : dict
        The extracted data as shaped parameters means, stds, weights, gammas
    """

    yvals = in_dist.objdata["yvals"]

    default = in_dist.metadata["xvals"][0]
    z = kwargs.pop("xvals", default)
    nz = kwargs.pop("nz", 300)

    minz = np.min(z)
    _, j = np.where(yvals > 0)
    maxz = np.max(z[j])
    newz = np.linspace(minz, maxz, nz)
    interp = sciinterp.interp1d(z, yvals, assume_sorted=True)
    newpdf = interp(newz)
    newpdf = newpdf / sciint.trapezoid(newpdf, newz).reshape(-1, 1)
    ALL, bigD, _ = build_sparse_representation(newz, newpdf)
    return dict(indices=ALL, metadata=bigD)



"""
Unit tests for PDF class
"""

import copy
import os
import unittest
from pathlib import Path
import tempfile
import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal
import scipy.stats as sps
import qp
from qp.plotting import init_matplotlib
from qp.parameterizations.packed_interp.packing_utils import PackingType
from tests.helpers.test_funcs import (
    assert_all_close,
    assert_all_small,
    build_ensemble,
)
from tests.helpers import test_data_helper as t_data


class EnsembleTestCase(unittest.TestCase):
    """Class to test qp.Ensemble functionality"""

    def setUp(self):
        """
        Make any objects that are used in multiple tests.
        """

        self.tmpdir = tempfile.TemporaryDirectory()

        self.norm_test_data = t_data.norm_test_data
        self.hist_test_data = t_data.hist_test_data
        self.interp_test_data = t_data.interp_test_data
        self.packed_interp_test_data = t_data.packed_interp_test_data

    def tearDown(self):
        "Clean up any mock data files created by the tests."

        self.tmpdir.cleanup()

    @staticmethod
    def _run_ensemble_funcs(ens: qp.Ensemble, xpts, tmpdir):
        """Run the test for a particular class"""

        pdfs = ens.pdf(xpts)
        cdfs = ens.cdf(xpts)
        logpdfs = ens.logpdf(xpts)
        logcdfs = ens.logcdf(xpts)

        if hasattr(ens.gen_obj, "npdf"):
            assert ens.npdf == ens.gen_obj.npdf

        with np.errstate(all="ignore"):
            assert np.allclose(np.log(pdfs), logpdfs, atol=1e-9)
            assert np.allclose(np.log(cdfs), logcdfs, atol=1e-9)

        binw = xpts[1:] - xpts[0:-1]
        check_cdf = ((pdfs[:, 0:-1] + pdfs[:, 1:]) * binw / 2).cumsum(axis=1) - cdfs[
            :, 1:
        ]
        assert_all_small(check_cdf, atol=5e-2, test_name="cdf")

        hist = ens.histogramize(xpts)[1]
        hist_check = ens.frozen.histogramize(xpts)[1]
        assert_all_small(hist - hist_check, atol=1e-5, test_name="hist")

        ppfs = ens.ppf(t_data.QUANTS)
        check_ppf = ens.cdf(ppfs) - t_data.QUANTS
        assert_all_small(check_ppf, atol=2e-2, test_name="ppf")

        sfs = ens.sf(xpts)
        check_sf = sfs + cdfs
        assert_all_small(check_sf - 1, atol=2e-2, test_name="sf")

        _ = ens.isf(t_data.QUANTS)
        check_isf = ens.cdf(ppfs) + t_data.QUANTS[::-1]
        assert_all_small(check_isf - 1, atol=2e-2, test_name="isf")

        samples = ens.rvs(size=1000)
        assert samples.shape[0] == ens.frozen.npdf
        assert samples.shape[1] == 1000

        median = ens.median()
        mean = ens.mean()
        var = ens.var()
        std = ens.std()
        entropy = ens.entropy()

        _ = ens.stats()
        modes = ens.mode(xpts)

        assert median.size == ens.npdf
        assert mean.size == ens.npdf
        assert np.std(mean) > 1e-8
        assert var.size == ens.npdf
        assert std.size == ens.npdf
        assert entropy.size == ens.npdf
        assert modes.size == ens.npdf

        integral = ens.integrate(limits=(ens.gen_obj.a, ens.gen_obj.a))
        interval = ens.interval(0.05)

        assert integral.size == ens.npdf
        assert interval[0].size == ens.npdf

        for N in range(3):
            moment_partial = ens.moment_partial(N, limits=(t_data.XMIN, t_data.XMAX))
            calc_moment = qp.metrics.calculate_moment(
                ens, N, limits=(t_data.XMIN, t_data.XMAX)
            )
            assert_all_close(
                moment_partial,
                calc_moment,
                rtol=5e-2,
                test_name="moment_partial_%i" % N,
            )

            sps_moment = ens.moment(N)
            assert sps_moment.size == ens.npdf
            # assert_all_close(sps_moment.flatten(), moment_partial.flatten(),
            #    rtol=5e-2, test_name="moment_%i" % N)
            # pmf = ens.pmf(N)
            # logpmf = ens.logpmf(N)

        init_matplotlib()
        axes = ens.plot(xlim=(xpts[0], xpts[-1]))
        ens.plot_native(axes=axes)

        red_ens = ens[np.arange(5)]
        red_pdf = red_ens.pdf(xpts)

        check_red = red_pdf - pdfs[0:5]
        assert_all_small(check_red, atol=1e-5, test_name="red")

        if hasattr(ens.gen_obj, "npdf"):  # skip scipy norm
            commList = [None]
            try:
                import mpi4py.MPI  # pylint: disable=import-outside-toplevel

                commList.append(
                    mpi4py.MPI.COMM_WORLD
                )  # pylint: disable=c-extension-no-member
            except ImportError:
                pass
            for comm in commList:
                filepath = Path(tmpdir) / "testwrite.hdf5"
                try:
                    group, fout = ens.initializeHdf5Write(filepath, ens.npdf, comm)
                except TypeError:
                    continue
                ens.writeHdf5Chunk(group, 0, ens.npdf)
                ens.finalizeHdf5Write(fout)
                readens = qp.read(filepath)
                assert readens.metadata.keys() == ens.metadata.keys()
                assert readens.objdata.keys() == ens.objdata.keys()
                # os.remove("testwrite.hdf5")

    @staticmethod
    def _run_merge_tests(ens: qp.Ensemble, xpts):
        npdf = ens.npdf
        pdf_orig = ens.pdf(xpts)

        ens_cat = qp.concatenate([ens, ens])
        ens.append(ens)

        pdf_cat = ens_cat.pdf(xpts)

        modes = np.array([xpts[idx] for idx in np.squeeze(np.argmax(pdf_cat, axis=1))])

        ens_cat.set_ancil({"mode": modes})
        pdf_app = ens.pdf(xpts)

        mask = np.concatenate([np.ones((npdf), "bool"), np.zeros((npdf), "bool")])
        ens_check = ens_cat[mask]
        pdf_check = ens_check.pdf(xpts)

        assert_all_close(pdf_cat, pdf_app, atol=5e-8, test_name="merge_1")
        assert_all_close(pdf_orig, pdf_check, atol=5e-8, test_name="merge_2")
        assert_all_close(
            ens_cat.ancil["mode"][mask], modes[mask], atol=5e-8, test_name="mode"
        )

    def test_norm(self):
        """Run the ensemble tests on an ensemble of scipy.stats.norm distributions"""
        key = "norm"
        # cls_test_data = qp.stats.norm_gen.test_data[key]  # pylint: disable=no-member
        cls_test_data = self.norm_test_data[key]
        ens_norm = build_ensemble(cls_test_data)
        assert hasattr(ens_norm, "gen_func")
        assert isinstance(
            ens_norm.gen_obj, qp.stats.norm_gen  # pylint: disable=no-member
        )
        assert "loc" in ens_norm.frozen.kwds
        self._run_ensemble_funcs(
            ens_norm, cls_test_data["test_xvals"], self.tmpdir.name
        )
        self._run_merge_tests(ens_norm, cls_test_data["test_xvals"])

    def test_hist(self):
        """Run the ensemble tests on an ensemble of qp.hist distributions"""
        key = "hist"

        # qp.hist_gen.make_test_data()
        # cls_test_data = qp.hist_gen.test_data[key]
        cls_test_data = self.hist_test_data[key]
        ens_h = build_ensemble(cls_test_data)
        assert isinstance(ens_h.gen_obj, qp.hist_gen)
        self._run_ensemble_funcs(ens_h, cls_test_data["test_xvals"], self.tmpdir.name)
        self._run_merge_tests(ens_h, cls_test_data["test_xvals"])

        pdfs_mod = copy.copy(ens_h.dist.pdfs)
        pdfs_mod[:, 7] = 0.5 * pdfs_mod[:, 7]
        ens_h.update_objdata(dict(pdfs=pdfs_mod))

    def test_interp(self):
        """Run the ensemble tests on an ensemble of qp.interp distributions"""
        key = "interp"
        # qp.interp_gen.make_test_data()
        # cls_test_data = qp.interp_gen.test_data[key]
        cls_test_data = self.interp_test_data[key]
        ens_i = build_ensemble(cls_test_data)
        assert isinstance(ens_i.gen_obj, qp.interp_gen)
        self._run_ensemble_funcs(ens_i, cls_test_data["test_xvals"], self.tmpdir.name)

    def test_packed_interp(self):
        """Run the ensemble tests on an ensemble of qp.packed_interp distributions"""
        key = "lin_packed_interp"

        # qp.packed_interp_gen.make_test_data()
        # cls_test_data = qp.packed_interp_gen.test_data[key]
        cls_test_data = self.packed_interp_test_data[key]
        ens_i = build_ensemble(cls_test_data)
        assert isinstance(ens_i.gen_obj, qp.packed_interp_gen)
        self._run_ensemble_funcs(ens_i, cls_test_data["test_xvals"], self.tmpdir.name)
        assert np.isfinite(ens_i.dist.yvals).all()

    def test_iterator(self):
        """Test the iterated read"""
        # QP_DIR = os.path.abspath(os.path.dirname(qp.__file__))
        TEST_DIR = Path(__file__).resolve().parent
        data_file = os.path.join(TEST_DIR.parent, "test_data", "test.hdf5")
        ens = qp.read(data_file)
        data_length = qp.data_length(data_file)
        assert data_length == ens.npdf
        itr = qp.iterator(data_file, 10)
        test_grid = np.linspace(0.0, 1.0, 11)
        for start, end, ens_i in itr:
            check_vals = ens[start:end].pdf(test_grid)
            test_vals = ens_i.pdf(test_grid)
            assert np.allclose(check_vals, test_vals)

    def test_mixmod_with_negative_weights(self):
        """Verify that an exception is raised when setting up a mixture model with negative weights"""
        means = np.array([0.5, 1.1, 2.9])
        sigmas = np.array([0.15, 0.13, 0.14])
        weights = np.array([1, 0.5, -0.25])
        with self.assertRaises(ValueError):
            _ = qp.mixmod(weights=weights, means=means, stds=sigmas)

    def test_dictionary_output(self):
        """Test that writing and reading a dictionary of ensembles works as expected."""
        key = "hist"
        # qp.hist_gen.make_test_data()
        # cls_test_data = qp.hist_gen.test_data[key]
        cls_test_data = self.hist_test_data[key]
        ens_h = build_ensemble(cls_test_data)

        key = "interp"
        # qp.interp_gen.make_test_data()
        # cls_test_data = qp.interp_gen.test_data[key]
        cls_test_data = self.interp_test_data[key]
        ens_i = build_ensemble(cls_test_data)

        output_dict = {
            "hist": ens_h,
            "interp": ens_i,
        }

        filepath = Path(self.tmpdir.name) / "test_dict.hdf5"
        qp.write_dict(filepath, output_dict)

        input_dict = qp.read_dict(filepath)

        assert input_dict.keys() == output_dict.keys()

        XVALS = np.linspace(0, 3, 100)
        for ens_type in ["hist", "interp"]:
            assert_array_equal(
                input_dict[ens_type].metadata["pdf_name"],
                output_dict[ens_type].metadata["pdf_name"],
            )

            assert_array_almost_equal(
                input_dict[ens_type].pdf(XVALS), output_dict[ens_type].pdf(XVALS)
            )


if __name__ == "__main__":
    unittest.main()



"""Integer packing utilities for qp"""

from __future__ import annotations

import enum

import numpy as np
from numpy.typing import ArrayLike


class PackingType(enum.Enum):
    linear_from_rowmax = 0
    log_from_rowmax = 1


def linear_pack_from_rowmax(input_array: ArrayLike) -> tuple[np.ndarray, np.ndarray]:
    """Pack an array into 8bit unsigned integers, using the maximum of each row as a reference

    This packs the values onto a linear grid for each row, running from 0 to row_max

    Parameters
    ----------
    input_array : ArrayLike
        The values we are packing

    Returns
    -------
    packed_array : np.ndarray
        The packed values
    row_max : np.ndarray
        The max for each row, need to unpack the array
    """
    row_max = np.expand_dims(input_array.max(axis=1), -1)
    return np.round(255 * input_array / row_max).astype(np.uint8), row_max


def linear_unpack_from_rowmax(
    packed_array: ArrayLike, row_max: ArrayLike
) -> np.ndarray[float]:
    """Unpack an array into 8bit unsigned integers, using the maximum of each row as a reference

    Parameters
    ----------
    packed_array : ArrayLike
        The packed values
    row_max : ArrayLike
        The max for each row, need to unpack the array


    Returns
    -------
    unpacked_array : np.ndarray[float]
        The unpacked values
    """
    unpacked_array = row_max * packed_array / 255.0
    return unpacked_array


def log_pack_from_rowmax(
    input_array: ArrayLike, log_floor: float = -3.0
) -> tuple[np.ndarray[np.uint8], np.ndarray]:
    """Pack an array into 8bit unsigned integers, using the maximum of each row as a reference

    This packs the values onto a log grid for each row, running from row_max / 10**log_floor to row_max

    Parameters
    ----------
    input_array : ArrayLike
        The values we are packing
    log_floor : float, optional
        The logarithmic floor used for the packing, by default -3.

    Returns
    -------
    packed_array : np.ndarray[np.uint8]
        The packed values
    row_max : np.ndarray
        The max for each row, need to unpack the array
    """
    neg_log_floor = -1.0 * log_floor
    epsilon = np.power(10.0, 3 * log_floor)
    row_max = np.expand_dims(input_array.max(axis=1), -1)
    return (
        np.round(
            255
            * (np.log10((input_array + epsilon) / row_max) + neg_log_floor)
            / neg_log_floor
        )
        .clip(0.0, 255.0)
        .astype(np.uint8),
        row_max,
    )


def log_unpack_from_rowmax(
    packed_array: ArrayLike, row_max: ArrayLike, log_floor: float = -3.0
) -> np.ndarray:
    """Unpack an array into 8bit unsigned integers, using the maximum of each row as a reference

    Parameters
    ----------
    packed_array : ArrayLike
        The packed values
    row_max : ArrayLike
        The max for each row, need to unpack the array
    log_floor : float, optional
        The logarithmic floor used for the packing, -3 by default.

    Returns
    -------
    unpacked_array : np.ndarray
        The unpacked values
    """
    neg_log_floor = -1.0 * log_floor
    unpacked_array = row_max * np.where(
        packed_array == 0,
        0.0,
        np.power(10, neg_log_floor * ((packed_array / 255.0) - 1.0)),
    )
    return unpacked_array


def pack_array(packing_type: PackingType, input_array: ArrayLike, **kwargs):
    """Pack an array into 8bit unsigned integers

    Parameters
    ----------
    packing_type : PackingType
        Enum specifying the type of packing to use
    input_array : ArrayLike
        The values we are packing
    kwargs
        depend on the packing type used

    Returns
    -------
    np.ndarray
        Details depend on packing type used
    """

    if packing_type == PackingType.linear_from_rowmax:
        return linear_pack_from_rowmax(input_array)
    if packing_type == PackingType.log_from_rowmax:
        return log_pack_from_rowmax(input_array, kwargs.get("log_floor", -3))
    raise ValueError(
        f"Packing for packing type {packing_type} is not implemented"
    )  # pragma: no cover


def unpack_array(packing_type: PackingType, packed_array: ArrayLike, **kwargs):
    """Unpack an array from 8bit unsigned integers

    Parameters
    ----------
    packing_type : PackingType
        Enum specifying the type of packing to use
    packed_array : ArrayLike
        The packed values
    kwargs
        depend on the packing type used

    Returns
    -------
    np.ndarray
        Details depend on packing type used
    """
    if packing_type == PackingType.linear_from_rowmax:
        return linear_unpack_from_rowmax(packed_array, row_max=kwargs.get("row_max"))
    if packing_type == PackingType.log_from_rowmax:
        return log_unpack_from_rowmax(
            packed_array,
            row_max=kwargs.get("row_max"),
            log_floor=kwargs.get("log_floor", -3),
        )
    raise ValueError(
        f"Unpacking for packing type {packing_type} is not implemented"
    )  # pragma: no cover



import pytest
from pathlib import Path
import numpy as np
import scipy.stats as sps
from tests.helpers import test_data_helper as t_data
import qp
from qp.parameterizations.packed_interp.packing_utils import PackingType


@pytest.fixture
def test_dir() -> Path:
    """Return path to test directory

    Returns
    -------
    Path
        Path to test directory
    """
    return Path(__file__).resolve().parent


@pytest.fixture
def test_data_dir(test_dir) -> Path:
    return test_dir / "test_data"


@pytest.fixture
def hist_test_data():
    """Writes out test data and parameters for the histogram parameterization"""

    return t_data.hist_test_data


@pytest.fixture
def interp_test_data():
    """Writes out test data and parameters for the interp parameterizations"""

    return t_data.interp_test_data


@pytest.fixture
def interp_irreg_test_data():

    return t_data.interp_irregular_test_data


@pytest.fixture
def mixmod_test_data():
    """Writes out test data and parameters for the mixmod parameterization"""

    return t_data.mixmod_test_data


@pytest.fixture
def quant_test_data():
    """Test data and parameters for the quant parameterization"""

    return t_data.quant_test_data


@pytest.fixture
def norm_test_data():
    """Writes out test data and parameters for the norm parameterization"""

    return t_data.norm_test_data


@pytest.fixture
def packed_interp_test_data():
    """Writes out test data and parameters for the packed interp parameterization"""

    return t_data.packed_interp_test_data


@pytest.fixture
def sparse_test_data():
    """Writes out test data and parameters for the sparse parameterization"""

    return t_data.sparse_test_data


@pytest.fixture
def spline_test_data():
    """Writes out test data and parameters for the spline parameterization"""

    return t_data.spline_test_data


## Fixtures to create ensembles
@pytest.fixture
def hist_ensemble(hist_test_data):
    ens_h = qp.hist.create_ensemble(**hist_test_data["hist"]["ctor_data"])
    return ens_h


@pytest.fixture
def norm_ensemble(norm_test_data):
    ens_n = qp.stats.norm.create_ensemble(norm_test_data["norm"]["ctor_data"])
    return ens_n



import pytest
import qp
import numpy as np

from tests.helpers import test_data_helper as t_data
from tests.helpers.test_funcs import assert_all_close


def test_repr(hist_ensemble):
    """Test that the representation method works as expected."""

    assert (
        hist_ensemble.__repr__()
        == f"Ensemble(the_class=hist,shape=({t_data.NPDF}, {t_data.NBIN -1}))"
    )


def test_len(hist_ensemble):
    """Make sure that the .__len__ method returns the number of distributions."""

    assert len(hist_ensemble) == hist_ensemble.npdf


@pytest.mark.parametrize(
    "ancil_data", [(np.array(["gal1", "gal2"])), (["test1", "test2"])]
)
def test_encode_decode_strings(tmp_path, ancil_data):
    """Tests that the encoding and decoding of string ancil data columns for HDF5 files works with
    write_to and read."""

    bins = np.linspace(-2, 2, 11)
    pdfs = np.array(
        [
            [0, 0.5, 1, 0.5, 0.5, 1.25, 1.5, 0.75, 0.5, 0.2],
            [0.05, 0.09, 0.15, 0.2, 0.3, 0.5, 0.25, 0.15, 0.1, 0.025],
        ]
    )
    ancil = {"names": ancil_data}
    ens_h = qp.hist.create_ensemble(bins, pdfs, ancil=ancil)

    # write to file
    file_path = tmp_path / "test-encode.hdf5"
    ens_h.write_to(file_path)

    # read from file
    new_ens = qp.read(file_path)
    assert new_ens.ancil["names"][0] == ancil_data[0]


# TODO: don't need this test?
@pytest.mark.parametrize(
    "loc,scale,npdf",
    [
        (np.array([1.0, 1.5]), np.array([1.0, 1.5]), 2),
        (np.array([[1], [0.5]]), np.array([[1], [0.5]]), 2),
    ],
)
def test_stats_arg_reshape(loc, scale, npdf):
    """Tests that the arg reshape and broadcasting for scipy classes works as expected"""

    ens_chi = qp.stats.chi2.create_ensemble(
        data={"scale": scale, "loc": loc, "df": 0.5}
    )

    assert ens_chi.npdf == npdf


def test_ensemble_objdata_dims(hist_test_data):
    """Ensure that the objdata arrays' dimensionality is reduced when slicing for 1 object."""

    key = "hist"
    ens_h = qp.Ensemble(
        hist_test_data[key]["gen_func"], hist_test_data[key]["ctor_data"]
    )

    assert np.ndim(ens_h[1].objdata["pdfs"]) == 1

    single_ens = ens_h[1]

    assert np.ndim(single_ens[0].objdata["pdfs"]) == 1
    with pytest.raises(IndexError) as exec_info:
        single_ens[1].objdata["pdfs"]
    assert exec_info.type is IndexError

    maxvals = np.max(hist_test_data[key]["ctor_data"]["pdfs"], axis=1)
    ancil = dict(maxvals=maxvals)
    ens_h.set_ancil(ancil)

    single_ens2 = ens_h[2]

    assert np.ndim(single_ens2[0].objdata["pdfs"]) == 1
    assert np.ndim(single_ens2[0].ancil["maxvals"]) == 0

    with pytest.raises(IndexError) as exec_info:
        single_ens2[2].ancil
    assert exec_info.type is IndexError


def test_norm_error_raised(norm_ensemble):
    """Make sure that the appropriate error is raised when a parameterizatoin doesn't have the normalize method."""

    with pytest.raises(
        AttributeError,
        match="This parameterization does not have a normalization function.",
    ):
        norm_ensemble.norm()


def test_x_samples(norm_ensemble):
    """Make sure that x_samples does as expected when the parameterization doesn't have an x_samples method."""

    min = 0
    max = 2
    n = 100
    x_samps = norm_ensemble.x_samples(min=min, max=max, n=n)

    assert_all_close(x_samps, np.linspace(min, max, n))


def test_functions_one_ensemble(hist_ensemble):
    """Make sure that the various Ensemble functions return the appropriate dimensions when given 1 distribution."""

    assert np.ndim(hist_ensemble[0].pdf(1.0)) == 0
    assert np.ndim(hist_ensemble[0].logpdf(1.0)) == 0
    assert np.ndim(hist_ensemble[0].cdf(1.0)) == 0
    assert np.ndim(hist_ensemble[0].logcdf(1.0)) == 0
    assert np.ndim(hist_ensemble[0].ppf(1.0)) == 0
    assert np.ndim(hist_ensemble[0].sf(1.0)) == 0
    assert np.ndim(hist_ensemble[0].logsf(1.0)) == 0
    assert np.ndim(hist_ensemble[0].isf(1.0)) == 0
    assert np.ndim(hist_ensemble[0].mean()) == 0
    assert np.ndim(hist_ensemble[0].median()) == 0
    assert np.ndim(hist_ensemble[0].std()) == 0
    assert np.ndim(hist_ensemble[0].var()) == 0
    assert np.ndim(hist_ensemble[0].moment(1)) == 0
    assert np.ndim(hist_ensemble[0].entropy()) == 0


def test_ancil_dimension(hist_ensemble):
    """Test that the ancillary data table dimensions make sense"""

    ancil = {"ids": np.ones((t_data.NPDF, 3))}
    hist_ensemble.set_ancil(ancil)
    assert_all_close(hist_ensemble.ancil["ids"], ancil["ids"])


def test_writeHdf5Chunk_single_ens(hist_ensemble, tmp_path):
    """Test that the writeHdf5Chunk function works with a chunk of 1 distribution."""

    single_ens = hist_ensemble[0]
    file_path = tmp_path / "test.hdf5"
    groups, fout = single_ens.initializeHdf5Write(file_path, single_ens.npdf)

    single_ens.writeHdf5Chunk(groups, 0, 1)
    single_ens.finalizeHdf5Write(fout)



from __future__ import annotations

from typing import List

import numpy as np

from .abstract_pdf_constructor import AbstractQuantilePdfConstructor
from .quant_utils import evaluate_hist_multi_x_multi_y


class PiecewiseConstant(AbstractQuantilePdfConstructor):
    """This constructor takes the input quantiles and locations, and calculates a numerical
    derivative. We assume a constant value between derivative points and interpolate between
    those.
    """

    def __init__(self, quantiles: List[float], locations: List[List[float]]) -> None:
        """Constructor to instantiate this class.

        Parameters
        ----------
        quantiles : List[float]
            List of n quantile values in the range (0,1).
        locations : List[List[float]]
            List of m Lists, each containing n values corresponding to the
            y-value of the PPF function at the same quantile index.
        """
        self._quantiles = quantiles
        self._locations = np.atleast_2d(locations)

        self._cdf_derivatives = None
        self._cdf_2nd_derivatives = None
        self._adjusted_locations = None

    def prepare_constructor(self) -> None:
        """This method will calculate the numerical derivative as well as the
        adjusted locations. The adjustments are necessary because the derivative is
        not a central derivative.
        """

        self._cdf_derivatives = np.zeros(self._locations.shape)
        self._cdf_2nd_derivatives = np.zeros(self._locations.shape)

        self._cdf_derivatives[:, 0:-1] = (
            self._quantiles[1:] - self._quantiles[0:-1]
        ) / (self._locations[:, 1:] - self._locations[:, 0:-1])
        self._cdf_2nd_derivatives[:, 0:-1] = (
            self._cdf_derivatives[:, 1:] - self._cdf_derivatives[:, 0:-1]
        )

        # Offset the locations by -(l_[i+1] - l_i) / 2. So that the cdf_deriv can be correctly located.
        # This offset is necessary to correctly place the _cdf_derivs because we are using a
        # forward difference to calculate the numerical derivative.
        self._adjusted_locations = self._locations[:, 1:] - np.diff(self._locations) / 2

    def construct_pdf(
        self, grid: List[float], row: List[int] = None
    ) -> List[List[float]]:
        """Take the intermediate calculations and return the interpolated y values
        given the input grid.

        Parameters
        ----------
        grid : List[float]
            List of x values to calculate y values for.
        row : List[int], optional
            A list of indexes of the original distribution to return, used as a filter.
            By default None will do no filtering.

        Returns
        -------
        List[List[float]]
            The lists of y values returned from self._interpolation_functions
        """
        if self._cdf_derivatives is None:
            self.prepare_constructor()

        # Theoretically, it should be possible for row to be passed in as None,
        # However the existing code that implements `interpolate_multi_x_multi_y`
        # doesn't handle `row = None`. So we hard code it here to be array([0]).
        if row is None:
            row = np.asarray([0])

        # ! The use of `.ravel` is legacy code. The original intent is unclear,
        # since it breaks the paradigm of maintaining the output as a List[List[float]].
        return evaluate_hist_multi_x_multi_y(
            grid,
            row,
            self._adjusted_locations,
            self._cdf_derivatives,
            self._cdf_2nd_derivatives,
        ).ravel()

    def debug(self):
        """Utility method to help with debugging. Returns input and intermediate
        calculations.

        Returns
        -------
            _quantiles :
                Input during constructor instantiation
            _locations :
                Input during constructor instantiation
            _cdf_derivatives :
                Numerical derivative using _quantiles and _locations
            _cdf_2nd_derivatives :
                Numerical second derivative using _quantiles and _locations
            _adjusted_locations :
                Result of shifting the locations due to the use of non-central numerical derivatives
        """
        return (
            self._quantiles,
            self._locations,
            self._cdf_derivatives,
            self._cdf_2nd_derivatives,
            self._adjusted_locations,
        )



"""This module implements a PDT distribution sub-class using splines"""

from __future__ import annotations
import numpy as np

from scipy.interpolate import splev, splint, splrep, interp1d
from scipy.special import errstate  # pylint: disable=no-name-in-module
from scipy.stats import rv_continuous
from typing import Mapping, Optional
from numpy.typing import ArrayLike

from .spline_utils import (
    extract_samples,
    spline_extract_xy_vals,
    build_kdes,
    evaluate_kdes,
    normalize_spline,
    build_splines,
)
from ...core.factory import add_class
from ...core.ensemble import Ensemble
from ..base import Pdf_rows_gen
from ...plotting import get_axes_and_xlims, plot_pdf_on_axes
from ...utils.array import reshape_to_pdf_size


class spline_gen(Pdf_rows_gen):
    """Spline based distribution

    Notes
    -----
    This implements PDFs using a set of splines

    The relevant data members are:

    - `splx`:  (npdf, n) spline-knot x-values
    - `sply`:  (npdf, n) spline-knot y-values
    - `spln`:  (npdf) spline-knot order parameters

    The pdf() for the ith pdf will return the result of
    `scipy.interpolate.splev(x, splx[i], sply[i], spln[i))`

    The cdf() for the ith pdf will return the result of
    `scipy.interpolate.splint(x, splx[i], sply[i], spln[i))`

    The ppf() will use the default scipy implementation, which
    inverts the cdf() as evaluated on an adaptive grid.
    """

    # pylint: disable=protected-access

    name = "spline"
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        """
        Create a new distribution using the given histogram

        Parameters
        --------
        splx : ArrayLike
          The x-values of the spline knots
        sply : ArrayLike
          The y-values of the spline knots
        spln : ArrayLike, optional
          The order of the spline knots, by default None

        Notes
        -----
        Either (splx, sply and spln) must be provided.
        """
        splx = kwargs.pop("splx", None)
        sply = kwargs.pop("sply", None)
        spln = kwargs.pop("spln", None)

        if splx is None:  # pragma: no cover
            raise ValueError("splx must be provided")
        if splx.shape != sply.shape:  # pragma: no cover
            raise ValueError(
                "Shape of xvals (%s) != shape of yvals (%s)" % (splx.shape, sply.shape)
            )
        # kwargs['a'] = self.a = np.min(splx)
        # kwargs['b'] = self.b = np.max(splx)
        self._xmin = np.min(splx)
        self._xmax = np.max(splx)
        # kwargs["shape"] = splx.shape[:-1]
        self._splx = reshape_to_pdf_size(splx, -1)
        self._sply = reshape_to_pdf_size(sply, -1)
        self._spln = reshape_to_pdf_size(spln, -1)

        kwargs["shape"] = self._splx.shape
        super().__init__(*args, **kwargs)
        self._addobjdata("splx", self._splx)
        self._addobjdata("sply", self._sply)
        self._addobjdata("spln", self._spln)

    @staticmethod
    def build_normed_splines(xvals, yvals, **kwargs):
        """
        Build a set of normalized splines using the x and y values

        Parameters
        ----------
        xvals : ArrayLike
          The x-values used to do the interpolation
        yvals : ArrayLike
          The y-values used to do the interpolation

        Returns
        -------
        splx : ArrayLike
          The x-values of the spline knots
        sply : ArrayLike
          The y-values of the spline knots
        spln : ArrayLike
          The order of the spline knots
        """
        if xvals.shape != yvals.shape:  # pragma: no cover
            raise ValueError(
                "Shape of xvals (%s) != shape of yvals (%s)"
                % (xvals.shape, yvals.shape)
            )
        xmin = np.min(xvals)
        xmax = np.max(xvals)

        # make sure xvals and yvals are 2d
        if np.ndim(xvals) == 1:
            xvals = np.expand_dims(xvals, axis=0)
        if np.ndim(yvals) == 1:
            yvals = np.expand_dims(yvals, axis=0)

        yvals = normalize_spline(xvals, yvals, limits=(xmin, xmax), **kwargs)
        return build_splines(xvals, yvals)

    @classmethod
    def create_from_xy_vals(cls, xvals, yvals, **kwargs):
        """
        Create a new distribution using the given x and y values

        Parameters
        ----------
        xvals : ArrayLike
          The x-values used to do the interpolation
        yvals : ArrayLike
          The y-values used to do the interpolation

        Returns
        -------
        pdf_obj : spline_gen
            The requested PDF
        """
        splx, sply, spln = spline_gen.build_normed_splines(xvals, yvals, **kwargs)
        gen_obj = cls(splx=splx, sply=sply, spln=spln)
        return gen_obj(**kwargs)

    @classmethod
    def create_from_samples(cls, xvals, samples, **kwargs):
        """
        Create a new distribution using the given x and y values

        Parameters
        ----------
        xvals : ArrayLike
          The x-values used to do the interpolation
        samples : ArrayLike
          The sample values used to build the KDE

        Returns
        -------
        pdf_obj : spline_gen
            The requested PDF
        """
        kdes = build_kdes(samples)
        kwargs.pop("yvals", None)
        yvals = evaluate_kdes(xvals, kdes)
        xvals_expand = (np.expand_dims(xvals, -1) * np.ones(samples.shape[0])).T
        return cls.create_from_xy_vals(xvals_expand, yvals, **kwargs)

    @property
    def splx(self) -> np.ndarray:
        """Return x-values of the spline knots"""
        return self._splx

    @property
    def sply(self) -> np.ndarray:
        """Return y-values of the spline knots"""
        return self._sply

    @property
    def spln(self) -> np.ndarray:
        """Return order of the spline knots"""
        return self._spln

    def _pdf(self, x, row):
        # pylint: disable=arguments-differ
        def pdf_row(xv, irow):
            return splev(
                xv, (self._splx[irow], self._sply[irow], self._spln[irow].item())
            )

        with errstate(all="ignore"):
            vv = np.vectorize(pdf_row)
        return vv(x, row).ravel()

    def _cdf(self, x, row):
        # pylint: disable=arguments-differ
        def cdf_row(xv, irow):
            return splint(
                self._xmin,
                xv,
                (self._splx[irow], self._sply[irow], self._spln[irow].item()),
            )

        with errstate(all="ignore"):
            vv = np.vectorize(cdf_row)
        return vv(x, row).ravel()

    def ppf(self, quants):
        # FIXME: remove this function once the issue with spline ppf is fixed
        raise NotImplementedError(
            "This function is buggy and currently not working properly, and will be restored once it's been fixed."
        )

    def _ppf(self, quants, row):
        # pylint: disable=arguments-differ

        # get the cdfs on a grid
        n_pts = 1001
        grid = np.linspace(self._xmin, self._xmax, n_pts)
        unique_rows = np.unique(row)
        cdf_vals = self._cdf(np.expand_dims(grid, -1), unique_rows).reshape(
            len(unique_rows), n_pts
        )

        def ppf_row(quantsv, irow):
            cdf_row = cdf_vals[irow]
            # Filter out the bits where it fluctuations down
            arg_sorted = np.argsort(cdf_row)
            sorted_vals = cdf_row[arg_sorted]
            sorted_grid = grid[arg_sorted]
            mask = np.zeros((len(sorted_grid)), dtype=bool)
            mask[1:] = sorted_grid[1:] > sorted_grid[0:-1]
            mask[1:] &= sorted_vals[1:] > sorted_vals[0:-1]
            sorted_masked_vals = sorted_vals[mask]
            sorted_masked_grid = sorted_grid[mask]
            sorted_masked_vals /= sorted_masked_vals[-1]
            # Build an interpolater, but reverse x and y to get the inverse function
            interp = interp1d(
                np.squeeze(sorted_vals[mask]),
                sorted_grid[mask],
                bounds_error=False,
                fill_value=(sorted_grid[0], sorted_grid[-1]),
            )
            return interp(quantsv)

        with errstate(all="ignore"):
            vv = np.vectorize(ppf_row)
        ret_vals = vv(quants, row).ravel()
        return ret_vals

    def _updated_ctor_param(self):
        """
        Set the bins as additional constructor argument
        """
        dct = super()._updated_ctor_param()
        dct["splx"] = self._splx
        dct["sply"] = self._sply
        dct["spln"] = self._spln
        return dct

    @classmethod
    def get_allocation_kwds(
        cls, npdf: int, **kwargs
    ) -> dict[str, tuple[tuple[int, int], str]]:
        """
        Return the keywords necessary to create an 'empty' hdf5 file with npdf entries
        for iterative file writeout.  We only need to allocate the objdata columns, as
        the metadata can be written when we finalize the file.

        Parameters
        ----------
        npdf : int
            number of *total* PDFs that will be written out
        kwargs
            dictionary of kwargs needed to create the ensemble

        Returns
        -------
        dict[str, tuple[tuple[int, int], str]]
        """
        if "splx" not in kwargs:  # pragma: no cover
            raise ValueError("required argument splx not included in kwargs")

        shape = np.shape(kwargs["splx"])
        return dict(splx=(shape, "f4"), sply=(shape, "f4"), spln=((shape[0], 1), "i4"))

    @classmethod
    def plot_native(cls, pdf, **kwargs):
        """Plot the PDF in a way that is particular to this type of distibution

        For a spline this shows the spline knots
        """
        axes, _, kw = get_axes_and_xlims(**kwargs)
        xvals = pdf.dist.splx[pdf.kwds["row"]]
        return plot_pdf_on_axes(axes, pdf, xvals, **kw)

    @classmethod
    def add_mappings(cls) -> None:
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)
        cls._add_creation_method(cls.create_from_xy_vals, "xy")
        cls._add_creation_method(cls.create_from_samples, "samples")
        cls._add_extraction_method(spline_extract_xy_vals, "xy")
        cls._add_extraction_method(extract_samples, "samples")

    @classmethod
    def create_ensemble(
        self,
        splx: ArrayLike,
        sply: ArrayLike,
        spln: Optional[ArrayLike] = None,
        ancil: Optional[Mapping] = None,
        method: Optional[str] = None,
    ) -> Ensemble:
        """Creates an Ensemble of distributions parameterized as via a set of splines.


        Parameters
        ----------
        splx : ArrayLike
          The x-values of the spline knots
        sply : ArrayLike
          The y-values of the spline knots
        spln : ArrayLike, optional
          The order of the spline knots, by default None
        ancil : Optional[Mapping], optional
            A dictionary of metadata for the distributions, where any arrays have the same length as the number of
            distributions, by default None
        method : Optional[str], optional
            The string of the creation method to use, by default None.

        Returns
        -------
        Ensemble
            An Ensemble object containing all of the given distributions.


        """
        data = {"splx": splx, "sply": sply, "spln": spln}
        return Ensemble(self, data, ancil, method)


spline = spline_gen
spline_from_xy = spline_gen.create_from_xy_vals
spline_from_samples = spline_gen.create_from_samples

add_class(spline_gen)



"""This module implements a PDT distribution sub-class using a Gaussian mixture model"""

from __future__ import annotations
import numpy as np
from scipy import stats as sps
from scipy.stats import rv_continuous
from typing import Mapping, Optional
from numpy.typing import ArrayLike
import warnings

from .mixmod_utils import extract_mixmod_fit_samples
from ...core.factory import add_class
from ..base import Pdf_rows_gen
from ...utils.array import (
    get_eval_case,
    reshape_to_pdf_size,
)
from ...utils.interpolation import interpolate_multi_x_y
from ...core.ensemble import Ensemble


class mixmod_gen(Pdf_rows_gen):
    """Parameterizes distributions using a Gaussian Mixture model.

    There are `ncomp` Gaussians in the model, and `npdf` distributions contained
    in the object.

    Parameters
    ----------
    means : ArrayLike
        The means of the Gaussians, with shape (npdf, ncomp)
    stds : ArrayLike
        The standard deviations of the Gaussians, with shape (npdf, ncomp)
    weights : ArrayLike
        The weights to attach to the Gaussians, with shape (npdf, ncomp).
        Weights should sum up to one. If not, the weights are interpreted
        as relative weights.
    warn : bool, optional
        If True, raises warnings if input is not finite. If False, no warnings
        are raised. By default True.


    Notes
    -----

    All distributions must have the same number of Gaussian components, `ncomp`.
    Use 0 as a fill value instead of `Nan`, which will result in errors in the
    PDF construction.


    Converting to this parameterization:

    This table contains the available methods to convert to this parameterization,
    their required arguments, and their method keys. If the key is `None`, this is
    the default conversion method.

    +------------------------------+--------------------------------------------+------------+
    | Function                     | Arguments                                  | Method key |
    +------------------------------+--------------------------------------------+------------+
    |`.extract_mixmod_fit_samples` | ncomps=3, nsamples=1000, random_state=None | None       |
    +------------------------------+--------------------------------------------+------------+

    Implementation Notes:

    The `pdf()` and `cdf()` are exact, and are computed as a weighted sum of
    the `pdf()` and `cdf()` of the component Gaussians.

    The `ppf()` is computed by computing the `cdf()` values on a fixed
    grid and interpolating the inverse function using `scipy.interp1d` with the
    default interpolation method (linear). `ppf(0)` returns negative infinity and
    `ppf(1)` returns positive infinity.


    """

    # pylint: disable=protected-access

    name = "mixmod"
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(
        self,
        means: ArrayLike,
        stds: ArrayLike,
        weights: ArrayLike,
        warn: bool = True,
        *args,
        **kwargs,
    ):
        """
        Create a new distribution or distributions using a Gaussian Mixture Model.
        There are ncomp Gaussians in the model, and npdf distributions contained
        in the object.

        Parameters
        ----------
        means : ArrayLike
            The means of the Gaussians, with shape (npdf, ncomp)
        stds : ArrayLike
            The standard deviations of the Gaussians, with shape (npdf, ncomp)
        weights : ArrayLike
            The weights to attach to the Gaussians, with shape (npdf, ncomp).
            Weights should sum up to one. If not, the weights are interpreted
            as relative weights.
        warn : bool, optional
            If True, raises warnings if input is not finite. If False, no warnings
            are raised. By default True.
        """
        self._scipy_version_warning()
        self._means = reshape_to_pdf_size(np.asarray(means), -1)
        self._stds = reshape_to_pdf_size(np.asarray(stds), -1)
        self._weights = reshape_to_pdf_size(np.asarray(weights), -1)
        kwargs["shape"] = self._means.shape  # means.shape
        self._ncomps = means.shape[-1]

        # validate input
        if (
            self._means.shape != self._stds.shape
            or self._stds.shape != self._weights.shape
        ):
            raise ValueError(
                f"Invalid input: means {np.shape(means)}, stds {np.shape(stds)}, and weights {np.shape(weights)} must have the same shape."
            )
        if np.any(self._weights < 0):
            raise ValueError(
                "Invalid input: All weights need to be larger than or equal to 0"
            )
        if np.any(self._stds < 0):
            raise ValueError(
                "Invalid input: All standard deviations (stds) must be greater than or equal to 0."
            )

        # raise warnings if input data is not finite
        self._warn = warn
        if self._warn:
            self._input_warnings()

        super().__init__(*args, **kwargs)

        self._weights = self._weights / self._weights.sum(axis=1)[:, None]
        self._addobjdata("weights", self._weights)
        self._addobjdata("stds", self._stds)
        self._addobjdata("means", self._means)

    def _scipy_version_warning(self):
        import scipy  # pylint: disable=import-outside-toplevel

        scipy_version = scipy.__version__
        vtuple = scipy_version.split(".")
        if int(vtuple[0]) > 1 or int(vtuple[1]) > 7:
            return
        raise DeprecationWarning(
            f"Mixmod_gen will not work correctly with scipy version < 1.8.0, you have {scipy_version}"
        )  # pragma: no cover

    def _input_warnings(self):
        """Raise warnings if input is not finite or any distribution has all stds <= 0"""
        if not np.all(np.isfinite(self._means)):
            indices = np.where(np.isfinite(self._means) != True)
            warnings.warn(
                f"The given means contain non-finite values for the following distributions: {indices}",
                RuntimeWarning,
            )
        if not np.all(np.isfinite(self._stds)):
            indices = np.where(np.isfinite(self._stds) != True)
            warnings.warn(
                f"There are non-finite values in the stds for the following distributions: {indices}",
                RuntimeWarning,
            )
        if not np.all(np.isfinite(self._weights)):
            indices = np.where(np.isfinite(self._weights) != True)
            warnings.warn(
                f"There are non-finite values in the weights for the following distributions: {indices}",
                RuntimeWarning,
            )
        if np.any(np.all(self._stds <= 0, axis=1)):
            indices = np.where(np.all(self._stds <= 0, axis=1))
            warnings.warn(
                f"The following distributions have all stds <= 0: {indices}",
                RuntimeWarning,
            )

    def normalize(self):
        raise RuntimeError(
            "The distributions in a mixmod parameterization are already normalized"
        )

    @property
    def weights(self) -> np.ndarray[float]:
        """Return weights to attach to the Gaussians"""
        return self._weights

    @property
    def means(self) -> np.ndarray[float]:
        """Return means of the Gaussians"""
        return self._means

    @property
    def stds(self) -> np.ndarray[float]:
        """Return standard deviations of the Gaussians"""
        return self._stds

    def x_samples(self) -> np.ndarray[float]:
        """Return a set of x values that can be used to plot all the PDFs."""

        # make sure the number of points is reasonable
        npts_min = 50
        npts_max = 10000

        # calculate bounds and dx
        dx = np.min(self._stds) / 2.0
        xmin = np.min(self._means) - np.max(self._stds)
        xmax = np.max(self._means) + np.max(self._stds)
        npts = (xmax - xmin) // dx
        if npts < npts_min:
            return np.linspace(xmin, xmax, npts_min)
        elif npts >= npts_min and npts <= npts_max:
            return np.linspace(xmin, xmax, int(npts))
        else:  # npts > npts_max
            return np.linspace(xmin, xmax, npts_max)

    def _pdf(self, x, row):
        # pylint: disable=arguments-differ
        if np.ndim(x) > 1:  # pragma: no cover
            x = np.expand_dims(x, -2)
        return (
            self.weights[row].swapaxes(-2, -1)
            * sps.norm(
                loc=self._means[row].swapaxes(-2, -1),
                scale=self._stds[row].swapaxes(-2, -1),
            ).pdf(x)
        ).sum(axis=0)

    def _cdf(self, x, row):
        # pylint: disable=arguments-differ
        if np.ndim(x) > 1:  # pragma: no cover
            x = np.expand_dims(x, -2)
        return (
            self.weights[row].swapaxes(-2, -1)
            * sps.norm(
                loc=self._means[row].swapaxes(-2, -1),
                scale=self._stds[row].swapaxes(-2, -1),
            ).cdf(x)
        ).sum(axis=0)

    def _ppf(self, x, row):
        # pylint: disable=arguments-differ
        min_val = np.min(self._means - 6 * self._stds)
        max_val = np.max(self._means + 6 * self._stds)
        grid = np.linspace(min_val, max_val, 201)
        case_idx, _, rr = get_eval_case(x, row)
        if case_idx == 1:
            cdf_vals = self.cdf(grid, rr)
        elif case_idx == 3:
            cdf_vals = self.cdf(grid, np.expand_dims(rr, -1))
        else:  # pragma: no cover
            raise ValueError(
                f"Oops, we can't handle this kind of input to mixmod._ppf {case_idx}"
            )
        return interpolate_multi_x_y(
            x, row, cdf_vals, grid, bounds_error=False, fill_value=(min_val, max_val)
        ).ravel()

    def _updated_ctor_param(self):
        """
        Set the bins as additional constructor argument
        """
        dct = super()._updated_ctor_param()
        dct["means"] = self._means
        dct["stds"] = self._stds
        dct["weights"] = self._weights
        dct["warn"] = self._warn
        return dct

    @classmethod
    def get_allocation_kwds(
        cls, npdf, **kwargs
    ) -> dict[str, tuple[tuple[int, int], str]]:
        """Return the kwds necessary to create an `empty` HDF5 file with ``npdf`` entries
        for iterative write. We only need to allocate the data columns, as
        the metadata will be written when we finalize the file.

        The number of data columns is calculated based on the length or shape of the
        metadata, ``n``. For example, the number of columns is ``nbins-1``
        for a histogram.

        Parameters
        ----------
        npdf : int
            Total number of distributions that will be written out
        kwargs :
            The keys needed to construct the shape of the data to be written.

        Returns
        -------
        dict[str, tuple[tuple[int, int], str]]
            A dictionary with a key for the objdata, a tuple with the shape of that data,
            and the data type of the data as a string.
            i.e. ``{objdata_key = ( (npdf, n), "f4" )}``

        Raises
        ------
        ValueError
            Raises an error if the means are not provided.
        """
        if "means" not in kwargs:  # pragma: no cover
            raise ValueError("Required argument `means` not included in kwargs")

        ncomp = np.shape(kwargs["means"])[-1]
        return dict(
            means=((npdf, ncomp), "f4"),
            stds=((npdf, ncomp), "f4"),
            weights=((npdf, ncomp), "f4"),
        )

    @classmethod
    def add_mappings(cls) -> None:
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)
        cls._add_extraction_method(extract_mixmod_fit_samples, None)

    @classmethod
    def create_ensemble(
        self,
        means: ArrayLike,
        stds: ArrayLike,
        weights: ArrayLike,
        warn: bool = True,
        ancil: Optional[Mapping] = None,
    ) -> Ensemble:
        """Creates an Ensemble of distributions parameterized as Gaussian Mixture models.

        `npdf` = the number of distributions

        `ncomp` = the number of Gaussians in the mixture model

        Parameters
        ----------
        means : ArrayLike
            The means of the Gaussians, with shape (npdf, ncomp)
        stds : ArrayLike
            The standard deviations of the Gaussians, with shape (npdf, ncomp)
        weights : ArrayLike
            The weights to attach to the Gaussians, with shape (npdf, ncomp).
            Weights should sum up to one. If not, the weights are interpreted
            as relative weights.
        warn : bool, optional
            If True, raises warnings if input is not finite. If False, no warnings
            are raised. By default True.
        ancil : Optional[Mapping], optional
            A dictionary of metadata for the distributions, where any arrays have
            the same length as the number of distributions, by default None


        Returns
        -------
        Ensemble
            An Ensemble object containing all of the given distributions.

        Examples
        --------

        To create an Ensemble of two distributions with associated ids:

        >>> import qp
        >>> means = np.array([[0.35, 0.55],[0.23,0.81]])
        >>> stds = np.array([[0.2, 0.25],[0.21, 0.19]])
        >>> weights = np.array([[0.4, 0.6],[0.3,0.7]])}
        >>> ancil = {'ids': [200, 205]}
        >>> ens = qp.mixmod.create_ensemble(means, stds, weights, ancil)
        >>> ens.metadata
        {'pdf_name': array([b'mixmod'], dtype='|S6'), 'pdf_version': array([0])}


        """
        data = {"means": means, "stds": stds, "weights": weights, "warn": warn}
        return Ensemble(self, data, ancil)


mixmod = mixmod_gen

add_class(mixmod_gen)



"""This is a template that developers can copy for creating new parameterizations"""

from __future__ import annotations
import numpy as np
from scipy.stats import rv_continuous
from typing import Mapping, Optional
from numpy.typing import ArrayLike
import warnings

# these imports will work when this file is placed in a folder in the parameterizations folder
from ..base import Pdf_rows_gen
from ...core.factory import add_class
from ...core.ensemble import Ensemble


# ---------------------------------------------------------------------
# How to use this template
# ---------------------------------------------------------------------
#
# This template is meant to provide you with a starting point when
# creating a new parameterization. In this template, square brackets like these
# [] indicate text that should be replaced with the relevant text for this
# parameterization. For example, [parameterization] indicates that you should
# replace the entire square brackets and text within with the name of the
# parameterization you are creating.
#
# The uncommented code provided within this template should
# not need to be altered except as directed in comments above the code,
# usually just indicating to replace variable names as necessary. The
# template also provides some code lines that can be uncommented and altered
# as instructed.
#
# Start by copying this template to a new folder named after your new
# parameterization type in the parameterizations folder. Within this folder,
# you should have the following files at a minimum:
# - __init__.py
# - [paramterization].py file
# - [parameterization]_utils.py


# -> Replace `parameterization` in the class name with your chosen parameterization name
#    and fill out the relevant information in the class docstring below
class parameterization_gen(Pdf_rows_gen):
    """[One sentence description goes here]

    [ any further description needed here]

    By default, the input distribution is normalized. If the input data is
    already normalized, you can use the optional parameter ``norm = False``
    to skip the normalization process.

    Parameters
    ----------

    Attributes
    ----------
    [arg1] : [dtype]
        [description]
    [arg2] : [dtype]
        [description]
    [warn] : bool, optional
            If True, raises warnings if input is not valid input data (i.e. if
            data is negative). If False, no warnings are raised. By default True.
    [[norm] : bool, optional
        If True, normalizes the input distribution. If False, assumes the
        given distribution is already normalized. By default True.]

    Methods
    -------
    create_ensemble(data,ancil)
        Create an ensemble with this parameterization.
    [add any additional optional methods here, i.e. plot_native]

    Notes
    -----

    Converting to this parameterization:

    This table contains the available methods to convert to this parameterization,
    their required arguments, and their method keys. If the key is `None`, this is
    the default conversion method.

    +----------+-----------+------------+
    | Function | Arguments | Method key |
    +----------+-----------+------------+
    |          |           |            |
    +----------+-----------+------------+

    Implementation notes:

    [any notes/caveats on how specific distributions are calculated]

    """

    # pylint: disable=protected-access

    name = "[parameterization]"
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(
        self, arg1: ArrayLike, arg2: ArrayLike, warn: bool = True, *args, **kwargs
    ):
        """
        Create a new distribution using the given data. [details]

        Parameters
        ----------
        [arg1] : ArrayLike
          [description]
        [arg2] : ArrayLike
          [description]
        warn : bool, optional
            If True, raises warnings if input is not valid input data (i.e. if
            data is negative). If False, no warnings are raised. By default True.
        [norm : bool, optional
            If True, normalizes the input distribution. If False, assumes the
            given distribution is already normalized. By default True.]
        """

        # ---------------------------------------------------------------------
        # The init function
        # ---------------------------------------------------------------------
        # -> Replace `arg1` and `arg2` in this function (and those throughout)
        #    with appropriate variable names for your coordinates and data arguments.
        #    Additional arguments can be added as necessary.
        # -> norm is an optional argument that you can include if the input data
        #    for this parameterization is a PDF (i.e. if it's normalizable.) If
        #    you include this parameter, add it to the init arguments and uncomment
        #    the optional normalize function below.

        # initialize the data
        self._arg1 = arg1
        self._arg2 = arg2

        # Set warn value and test for warnings as necessary
        self._warn = warn
        # -> Change relevant values below and add any additional warnings as necessary
        if self._warn:
            if not np.all(np.isfinite(self._arg1)):
                warnings.warn(
                    f"The given [arg1] contain non-finite values - {self._arg1}",
                    RuntimeWarning,
                )
            if not np.all(np.isfinite(self._arg2)):
                indices = np.where(np.isfinite(self._arg2) != True)
                warnings.warn(
                    f"There are non-finite values in [arg2] for the distributions: {indices[0]}",
                    RuntimeWarning,
                )

        # -> Uncomment code below to check if the input data is normalized
        # self._norm = norm
        # if self._norm:
        #     self._arg2 = self.normalize()

        # Get the shape of the data
        # and pass it to the base constructor to set up other attributes
        kwargs["shape"] = self._arg2.shape
        super().__init__(*args, **kwargs)

        # define data and metadata
        # metadata is shared across all distributions
        # data is quantities defined for each distribution
        self._addmetadata("arg1", self._arg1)
        self._addobjdata("arg2", self._arg2)

    # ---------------------------------------------------------------------
    # Functions that allow access to the data as attributes
    # ---------------------------------------------------------------------
    # Here we have properties for the two arguments, `arg1` and `arg2` that
    # we created. Add more functions if your parameterization has more
    # arguments.

    # property that allows access to the 'metadata' field
    # -> Replace `arg1` with the appropriate metadata variable
    @property
    def arg1(self):
        """Return arg1"""
        return self._arg1

    # property that allows access to the 'data' field
    # -> Replace `arg2` with the appropriate data variable
    @property
    def arg2(self):
        """Return arg2"""
        return self._arg2

    def x_samples(self):
        """Return a set of x values that can be used to plot all the [distributions]."""
        # This function is meant to return a set of x values that can be used to calculate
        # either the PDF or CDF, whichever is the characteristic distribution of
        # this parameterization. This means that the x values should allow you to plot
        # the data in the parameterization without missing anything due to too
        # large spacing between x values, for example.

        # -> Add in functionality to return an array of x values here
        return

    def _pdf(self, x, row):
        # pylint: disable=arguments-differ

        # ---------------------------------------------------------------------
        # The pdf evaluation function
        # ---------------------------------------------------------------------
        # This typically takes the form of a function from the
        # [paramterization]_utils.py file, where the output
        # is typically a 2D array of shape (npdf, n)
        # .ravel() is necessary to ensure that the array is 1D
        # when passed into scipy's rv_continuous pdf function

        # -> Uncomment the line below and replace the function with
        #    the relevant function and arguments, then delete the additional
        #    return statement
        #
        # return function_to_evaluate_pdf.ravel()
        return

    def _cdf(self, x, row):
        # pylint: disable=arguments-differ

        # ---------------------------------------------------------------------
        # The cdf evaluation function
        # ---------------------------------------------------------------------
        # This typically takes the form of a function from the
        # [parameterization]_utils.py file, where the output
        # is typically a 2D array of shape (npdf, n)
        # .ravel() is necessary to ensure that the array is 1D
        # when passed into scipy's rv_continuous cdf function

        # -> Uncomment the line below and replace the function with
        #    the relevant function and arguments, then delete the additional
        #    return statement
        #
        # return function_to_evaluate_cdf.ravel()
        return

    def _updated_ctor_param(self) -> Mapping:
        """
        Sets the arguments as additional constructor arguments. This function is needed
        by scipy in order to copy distributions, and makes a dictionary of all parameters
        necessary to construct the distribution.
        """
        dct = super()._updated_ctor_param()

        # -> replace `arg1` and `arg2` with your arguments and add any additional ones necessary
        dct["arg1"] = self._arg1
        dct["arg2"] = self._arg2
        dct["warn"] = self._warn
        return dct

    @classmethod
    def add_mappings(cls):
        """
        Adds this class' mappings to the conversion dictionary. Specifically, this should include at
        least a creation method and a function to extract the necessary values from the distribution
        to provide to the creation method. Together these functions should allow conversion from any
        other parameterization to this one.

        """
        # ---------------------------------------------------------------------
        # Add the creation function
        # ---------------------------------------------------------------------
        # This should always be cls.create, and with a key of `None` to make it
        # the default.
        cls._add_creation_method(cls.create, None)

        # You may add additional creation methods here, but they need to have a
        # specific key so they can be referred to when converting
        #
        # -> Uncomment the line below to add additional creation method:
        #    cls._add_creation_method(creation_method_function, "method_key")

        # ---------------------------------------------------------------------
        # Add the extraction function(s)
        # ---------------------------------------------------------------------
        #
        # At least one extraction method is required. The key for this
        # extraction method should be `None` to make it the default.

        # -> To add an extraction method, uncomment the line of code below
        #    and change `func_to_convert_default` to a useable function,
        #    either in parameterization_utils.py, or in utils.conversion.py
        #
        # cls._add_extraction_method(func_to_convert_default, None)

        # You can optionally provide additional conversion methods
        # These need to have a specific key so the user can refer to
        # them when converting.
        # -> Uncomment the line below and change the function and key
        #    to appropriate values
        #
        # cls._add_extraction_method(other_func_to_convert, "method_key")

    @classmethod
    def get_allocation_kwds(cls, npdf: int, **kwargs) -> Mapping:
        """Return the kwds necessary to create an `empty` HDF5 file with ``npdf`` entries
        for iterative write. We only need to allocate the data columns, as
        the metadata will be written when we finalize the file.

        The number of data columns is calculated based on the length or shape of the
        metadata, ``n``. For example, the number of columns is ``nbins-1``
        for a histogram.

        Parameters
        ----------
        npdf : int
            Total number of distributions that will be written out
        kwargs :
            The keys needed to construct the shape of the data to be written.

        Returns
        -------
        Mapping
            A dictionary with a key for the objdata, a tuple with the shape of that data,
            and the data type of the data as a string.
            i.e. ``{objdata_key = ( (npdf, n), "f4" )}``

        Raises
        ------
        ValueError
            Raises an error if the required kwarg is not provided.
        """

        # Checks that the required keyword is present in the function call
        # -> In the lines below fill out `arg1` with the appropriate
        #    metadata key
        if "arg1" not in kwargs:
            raise ValueError("required argument 'arg1' not included in kwargs")

        # Gets the shape of the coordinates/metadata. The line below is a
        # sample of how to do it, but you may need to modify as necessary.
        # -> Replace `arg1` and rename `narg1` below
        narg1 = len(kwargs["arg1"].flatten())
        return dict(pdfs=((npdf, narg1), "f4"))

    @classmethod
    def create_ensemble(
        self,
        arg1: ArrayLike,
        arg2: ArrayLike,
        warn: bool = True,
        ancil: Optional[Mapping] = None,
    ) -> Ensemble:
        """Creates an Ensemble of distributions parameterized as [parameterization].


        Parameters
        ----------
        [arg1] : ArrayLike
          [description]
        [arg2] : ArrayLike
          [description]
        [warn] : bool, optional
            If True, raises warnings if input is not valid input data (i.e. if
            data is negative). If False, no warnings are raised. By default True.
        ancil : Optional[Mapping], optional
            A dictionary of metadata for the distributions, where any arrays have the
            same length as the number of distributions, by default None

        Returns
        -------
        Ensemble
            An Ensemble object containing all of the given distributions.

        Examples
        --------

        To create an Ensemble with two distributions and an 'ancil' table that provides ids for the distributions, you can use the following code:

        >>> import qp
        >>> ancil = {'ids': [test ids] }
        >>> ens = qp.[parameterization].create_ensemble(arg1, arg2,ancil=ancil)
        >>> ens.metadata()
        [output here]

        """

        # pack up the data in to a dictionary to be passed to Ensemble
        # -> add norm below, in the docstring, and in the function call if your
        #    parameterization has a normalization function
        data = {"arg1": arg1, "arg2": arg2, "warn": warn}
        return Ensemble(self, data, ancil)

    # ---------------------------------------------------------------------
    # Optional normalization method
    # ---------------------------------------------------------------------
    # This optional method is meant to be used to normalize the distribution
    # data both when initializing the Ensemble and after it has been initialized.
    # -> If your parameterization input data can be normalized, uncomment the
    #    block of code below and add code to normalize the data (in this template
    #    arg2).
    # def normalize(self):

    #     # -> add some operations on object data to normalize here
    #     normed = self._arg2
    #     return normed

    # ---------------------------------------------------------------------
    # Optional methods
    # ---------------------------------------------------------------------
    # These are methods that will be calculated using scipy's underlying
    # functions, if no method is included for them here. You should only
    # add a method for one of these functions if it will speed up the
    # calculations or otherwise improve upon scipy's functionality.

    # -> Uncomment any method below if using

    # def _ppf():
    #     Keep in mind that by default ppf(0) and ppf(1) will not call this function
    #     and will instead return negative and positive infinity, respectively.
    #     pass

    # def _sf():
    #     pass

    # def _isf():
    #     pass

    # def _rvs():
    #     pass

    # ---------------------------------------------------------------------
    # Optional plotting method
    # ---------------------------------------------------------------------
    # Set a native plotting method that allows users to quickly plot PDFs
    # of your parameterization.
    # -> Uncomment the method below if used

    # @classmethod
    # def plot_native(cls, pdf, **kwargs):
    #     """Plot the PDF in a way that is particular to this type of distribution.
    #     [specific description of the plot for this parameterization]
    #
    #     Parameters
    #     ----------
    #     """
    #     # -> Add any plotting functions you create for this to the
    #     #    `plotting.py`` file in the `utils`` folder.

    #     # You should return the axes with the plot already created on them
    #     # as the return statement of the function.
    #     pass


# ---------------------------------------------------------------------
# Set alias and add class to factory
# ---------------------------------------------------------------------
# -> Replace `parameterization` in the lines below with the name of your parameterization
parameterization = parameterization_gen

# Register the class with the factory
# -> Replace `parameterization` with the name of your parameterization
add_class(parameterization_gen)



import pytest
import numpy as np
import warnings

import qp
from tests.helpers.test_data_helper import NPDF
from tests.helpers.test_funcs import assert_all_close


@pytest.fixture
def interp_ensemble(interp_test_data):
    ens_i = qp.interp.create_ensemble(**interp_test_data["interp"]["ctor_data"])
    return ens_i


@pytest.fixture
def irreg_interp_ensemble(interp_irreg_test_data):
    ens_i = qp.interp_irregular.create_ensemble(
        **interp_irreg_test_data["interp_irregular"]["ctor_data"]
    )
    return ens_i


#
# Interp Ensemble tests
#


def test_norm(interp_test_data):
    """Test that the norm method works as expected."""

    norm = False
    ancil = {"ancil": np.linspace(0, 11, NPDF)}
    ens_i = qp.interp.create_ensemble(
        **interp_test_data["interp"]["ctor_data"], norm=norm, ancil=ancil
    )

    assert ens_i.npdf == NPDF

    # make sure that values were not normalized
    xmin = ens_i.frozen.dist._xmin
    xmax = ens_i.frozen.dist._xmax
    integral = ens_i.integrate([xmin, xmax])
    assert not np.allclose(integral, np.ones(NPDF), rtol=1e-2)

    # make sure that the normalization has worked as expected
    ens_i.norm()
    integral_norm = ens_i.integrate([xmin, xmax])
    assert_all_close(integral_norm, np.ones(NPDF), rtol=1e-2)

    # make sure that the ancil table has not changed
    assert ens_i.ancil == ancil


def test_xsamples(interp_ensemble):
    """Make sure that x_samples generates values as expected."""

    x_samps = interp_ensemble.x_samples()
    assert len(interp_ensemble.metadata["xvals"]) == len(x_samps)

    assert_all_close(x_samps, interp_ensemble.metadata["xvals"])


@pytest.mark.parametrize(
    "xvals, yvals, match_string",
    [
        (
            np.array([0, 0.5, np.inf]),
            np.array([1.0, 2.0, 3.0]),
            "The given xvals contain non-finite values",
        ),
        (
            np.array([0, 0.5, 1]),
            np.array([1, np.nan, 1]),
            "There are non-finite values in the yvals for the following distributions:",
        ),
        (
            np.array([0, 0.5, 1]),
            np.array([-0.5, 1, 1.5]),
            "There are negative values in the yvals for the following distributions:",
        ),
    ],
)
def test_warnings(xvals, yvals, match_string):
    """Test that warnings when building an interpolated Ensemble work."""

    with pytest.warns(RuntimeWarning, match=match_string):
        ens_i = qp.interp.create_ensemble(xvals=xvals, yvals=yvals)


@pytest.mark.parametrize(
    "xvals, yvals, match_string",
    [
        (
            np.array([1.0, 1.5, 2.0]),
            np.array([1, 2, 3, 4]),
            "Shape of xbins in xvals",
        ),
        (
            np.array([0.5, 1.5, 1.0]),
            np.array([1, 3, 2]),
            "Invalid xvals: The given xvals are not sorted:",
        ),
        (
            np.array([0, 0.5, 1]),
            np.array([-0.5, -1, -0.5]),
            "cannot be properly normalized",
        ),
    ],
)
def test_invalid_input(xvals, yvals, match_string):
    """Tests that the appropriate errors are raised when invalid input is given"""

    with pytest.raises(ValueError, match=match_string) as err:
        ens_h = qp.interp.create_ensemble(xvals=xvals, yvals=yvals, warn=False)


#
# Irregular Interp Ensemble tests
#


def test_irreg_norm(interp_irreg_test_data):
    """Test that the norm method works as expected."""

    norm = False
    ancil = {"ancil": np.linspace(0, 11, NPDF)}
    ens_i = qp.interp_irregular.create_ensemble(
        **interp_irreg_test_data["interp_irregular"]["ctor_data"],
        norm=norm,
        ancil=ancil,
    )

    assert ens_i.npdf == NPDF

    # make sure that values were not normalized
    xmin = ens_i.frozen.dist._xmin
    xmax = ens_i.frozen.dist._xmax
    integral = ens_i.integrate([xmin, xmax])
    assert not np.allclose(integral, np.ones(NPDF), rtol=1e-2)

    # make sure that the normalization has worked as expected
    ens_i.norm()
    integral_norm = ens_i.integrate([xmin, xmax])
    assert_all_close(integral_norm, np.ones(NPDF), rtol=1e-2)

    # make sure that the ancil table has not changed
    assert ens_i.ancil == ancil


def test_irreg_xsamples(irreg_interp_ensemble):
    """Make sure that x_samples generates values as expected."""

    x_samps = irreg_interp_ensemble.x_samples()
    dx = np.min(np.diff(irreg_interp_ensemble.frozen.dist._yvals))
    compar_val = np.arange(
        np.min(irreg_interp_ensemble.frozen.dist._xvals),
        np.max(irreg_interp_ensemble.frozen.dist._yvals),
        dx,
    )

    assert_all_close(x_samps, compar_val)


@pytest.mark.parametrize(
    "xvals, yvals, match_string",
    [
        (
            np.array([0, 0.5, np.inf]),
            np.array([1.0, 2.0, 3.0]),
            "The given xvals contain non-finite values for the following distributions",
        ),
        (
            np.array([0, 0.5, 1]),
            np.array([1, np.nan, 1]),
            "There are non-finite values in the yvals for the following distributions:",
        ),
        (
            np.array([0, 0.5, 1]),
            np.array([-0.5, 1, 1.5]),
            "There are negative values in the yvals for the following distributions:",
        ),
    ],
)
def test_irreg_warnings(xvals, yvals, match_string):
    """Test that warnings when building an interpolated Ensemble work."""

    with pytest.warns(RuntimeWarning, match=match_string):
        ens_i = qp.interp_irregular.create_ensemble(xvals=xvals, yvals=yvals)


@pytest.mark.parametrize(
    "xvals, yvals, match_string",
    [
        (
            np.array([1.0, 1.5, 2.0]),
            np.array([1, 2, 3, 4]),
            "Shape of xvals",
        ),
        (
            np.array([0.5, 1.5, 1.0]),
            np.array([1, 3, 2]),
            "Invalid xvals: The given xvals are not sorted:",
        ),
        (
            np.array([0, 0.5, 1]),
            np.array([-0.5, -1, -0.5]),
            "The integral is < 0 for distributions at indices",
        ),
    ],
)
def test_irreg_invalid_input(xvals, yvals, match_string):
    """Tests that the appropriate errors are raised when invalid input is given"""

    with pytest.raises(ValueError, match=match_string) as err:
        ens_i = qp.interp_irregular.create_ensemble(
            xvals=xvals, yvals=yvals, warn=False
        )


def test_compute_ycumul_error():
    """Make sure that computing the CDF when it's less than 0 raises the appropriate error, even when norm=False."""

    xvals = np.array([0, 0.5, 1])
    yvals = np.array([-0.5, -1, -0.5])
    ens_i = qp.interp_irregular.create_ensemble(
        xvals=xvals, yvals=yvals, norm=False, warn=False
    )
    with pytest.raises(
        ValueError, match="The integral is < 0 for distributions at indices"
    ):
        value = ens_i.cdf(1.0)



import sys
import numpy as np

epsilon = sys.float_info.epsilon


def safelog(arr, threshold=epsilon):
    """
    Takes the natural logarithm of an array of potentially non-positive numbers

    Parameters
    ----------
    arr: numpy.ndarray, float
        values to be logged
    threshold: float
        small, positive value to replace zeros and negative numbers

    Returns
    -------
    logged: numpy.ndarray
        logarithms, with approximation in place of zeros and negative numbers
    """
    return np.log(np.array(arr).clip(threshold, np.inf))
